"""
Provides a class ``hash_machine`` encapsulating the two basic hash utilities used accross the library.
Instances of this class should receive their configuration parameters from the ``ENCODINGS`` and
``HASH_TYPES`` global variables of this module
"""

import pymerkle.encodings  # Load encoding types
import hashlib

ENCODINGS = pymerkle.encodings.ENCODINGS
"""Supported encoding types"""
HASH_TYPES = [
    'md5',
    'sha224',
    'sha256',
    'sha384',
    'sha512',
    'sha3_224',
    'sha3_256',
    'sha3_384',
    'sha3_512'
]
"""Supported hash types"""


class hash_machine(object):
    """Encapsulates the two basic hash utilities used accross this library.

    Sole purpose of this class is to fix at construction the hash and encoding types used for encryption,
    so that these parameters need not be redefined every time a hash utility is invoked. Instances
    of this class are thus to be initialized with every new construction of a Merkle-tree or every time
    a proof validation is about to be performed

    :param hash_type: specifies the hash algorithm to be used by the machine; must be among the elements of the ``HASH_TYPES``
                      global variable (upper- or mixed-case with '-' instead of '_' allowed). Defaults to ``'sha256'``
                      if unspecified
    :type hash_type:  str
    :param encoding:  specifies the encoding algorithm to be used by machine before hashing; must be among the elements of the
                      ENCODINGS global variable (upper- or mixed-case with ``-`` instead of ``_`` allowed). Defaults to ``'utf_8'``
                      if unspecified
    :type encoding:   str
    :param security:  defaults to ``True``, in which case security standards are applied against second-preimage attack, i.e.,
                      single, resp. double arguments of the ``.hash`` method will be prepended with ``'\\x00'``, resp. ``'\\x01'``
                      before hashing
    :type security:   bool

    :raises Exception: if ``hash_type``, resp. ``encoding`` is not contained in ``hashing.HASH_TYPES``, resp. ``hashing.ENCODINGS``

    :ivar HASH_ALGORITHM: (*builtin_function_or_method*) Hash algorithm used by the machine, specified in the obvious way
                          by the ``hash_type`` argument at construction
    :ivar ENCODING:       (*str*) Encoding type used by the machine before hashing, specified in the obvious way
                          by the ``encoding`` argument at construction
    :ivar SECURITY:       (*bool*) Indicates whether defense against second-preimage attack is activated, specified in the
                          obvious way by the ``security`` argument at construction
    """

    # -------------------------------- Construction --------------------------

    def __init__(self, hash_type='sha256', encoding='utf-8', security=True):

        # Hash algorithm used by the machine
        self.HASH_ALGORITHM = hash_machine.select_hash_algorithm(
            hash_type=hash_type.lower().replace('-', '_'))

        # Encoding type used by this machine before hashing
        self.ENCODING = hash_machine.select_encoding(
            encoding=encoding.lower().replace('-', '_'))

        self.SECURITY = security
        if self.SECURITY:
            # Prefices will be prepended before hashing for defense against
            # second-preimage attack
            self.PREFIX_0, self.PREFIX_1 = '\x00', '\x01'

    @staticmethod
    def select_hash_algorithm(hash_type):
        """Extracts from the ``hashlib`` module the desired hash algorithm according to the inserted label

        :param hash_type: label indicating the desired hash algorithm
        :type hash_type:  str
        :returns:         the desired hash algorithm; e.g. ``hashlib.sha256``
        :rtype:           builtin_function_or_method

        :raises Exception: if ``hash_type`` is not contained in ``hashing.HASH_TYPES``
        """
        if hash_type in HASH_TYPES:
            return getattr(hashlib, hash_type)
        else:
            message = f'\n\n * Hash type {hash_type} is not supported\n'
            raise Exception(message)

    @staticmethod
    def select_encoding(encoding):
        """Sole purpose of this function is to throw an exception if the inserted encoding type is not supported

        :param encoding: label indicating the desired encoding type
        :type encoding:  str
        :returns:        just the inserted argument
        :rtype:          str

        :raises Exception: if ``encoding`` is not contained in ``hashing.ENCODINGS``
        """
        if encoding in ENCODINGS:
            return encoding
        else:
            raise Exception(
                '\n\n * Encoding type {encoding} is not supported\n'.format(encoding=encoding))

    # ------------------------------- Hash utils -----------------------------

    def hash(self, first, second=None):
        """Core hash utility

        Returns the hash of the object occuring by concatenation of arguments in the given order.
        If only one argument is passed in, then the hash of this single argument is returned

        :param first:  left member of the pair to be hashed
        :type first:   str or bytes or bytearray
        :param second: [optional] right member of the pair to be hashed
        :type second:  bytes or bytearray

        .. warning:: if ``second`` is provided, then ``first`` *must* also be of `bytes`
                    or `byetarray` type

        :returns:      the hash of the provided pair
        :rtype:        bytes
        """

        if not second:  # one arg case

            if isinstance(first, (bytes, bytearray)):
                # bytes-like input

                if self.SECURITY:
                    # Apply security stadards
                    hex_hash = self.HASH_ALGORITHM(
                        bytes(
                            self.PREFIX_0,
                            encoding=self.ENCODING) +
                        first).hexdigest()
                    return bytes(hex_hash, encoding=self.ENCODING)

                # No security standards
                hex_hash = self.HASH_ALGORITHM(first).hexdigest()
                return bytes(hex_hash, encoding=self.ENCODING)

            # Non bytes-like input

            if self.SECURITY:

                # Apply security standards
                hex_hash = self.HASH_ALGORITHM(
                    bytes(
                        '{}{}'.format(
                            self.PREFIX_0,
                            first),
                        encoding=self.ENCODING)).hexdigest()
                return bytes(hex_hash, encoding=self.ENCODING)

            # No security standards
            hex_hash = self.HASH_ALGORITHM(
                bytes(first, encoding=self.ENCODING)).hexdigest()
            return bytes(hex_hash, encoding=self.ENCODING)

        # two args case

        if self.SECURITY:

            # Apply security standards
            hex_hash = self.HASH_ALGORITHM(
                bytes(
                    '{}{}{}{}'.format(
                        self.PREFIX_1,
                        first.decode(encoding=self.ENCODING),
                        self.PREFIX_1,
                        second.decode(encoding=self.ENCODING)),
                    encoding=self.ENCODING)).hexdigest()
            return bytes(hex_hash, encoding=self.ENCODING)

        # No security standards
        hex_hash = self.HASH_ALGORITHM(
            bytes(
                '{}{}'.format(
                    first.decode(encoding=self.ENCODING),
                    second.decode(encoding=self.ENCODING)),
                encoding=self.ENCODING)).hexdigest()
        return bytes(hex_hash, encoding=self.ENCODING)

    def multi_hash(self, signed_hashes, start):
        """Hash utility used in proof validation

        Repeatedly applies the ``.hash`` method over a tuple of signed hashes parenthesized in pairs
        as specified by accompanying signs. Schematically speaking, the result of

            ``multi_hash(signed_hashes=((1, a), (1, b), (-1, c), (-1, d)), start=1)``

        is equivalent to ``hash(hash(a, hash(b, c)), d)``. If the given sequence of signed hashes
        contains only one member, then this one hash is returned (without sign). That is,

            ``multi_hash(signed_hashes=((+/-1, a)), start=1)``

        is the same as ``a`` (no hashing over unique elements)
        .. warning:: When using this method, make sure that the combination of signs corresponds indeed
                     to a valid parenthetization

        :param signed_hashes: a sequence of signed hashes
        :type signed_hashes:  tuple of (+1/-1, bytes) pairs
        :param start:         position where the application of ``.hash`` will start from
        :type start:          int
        :returns:             the computed hash
        :rtype:               bytes

        .. note:: Returns ``None`` if the inserted sequence of signed hashes was empty
        """

        signed_hashes = list(signed_hashes)

        # Calculate and return
        if signed_hashes != []:
            if len(signed_hashes) > 1:
                i = start
                while len(signed_hashes) > 1:
                    if signed_hashes[i][0] == +1:
                        # Pair with the right neighbour
                        if i == 0:
                            new_sign = +1
                        else:
                            new_sign = signed_hashes[i + 1][0]
                        new_hash = self.hash(
                            signed_hashes[i][1], signed_hashes[i + 1][1])
                        move = +1
                    else:
                        # Pair with the left neighbour
                        new_sign = signed_hashes[i - 1][0]
                        new_hash = self.hash(
                            signed_hashes[i - 1][1], signed_hashes[i][1])
                        move = -1
                    # Store and shrink
                    signed_hashes[i] = (new_sign, new_hash)
                    del signed_hashes[i + move]
                    if move < 0:
                        i -= 1
                return signed_hashes[0][1]
            # signed_hashes contained one element
            return signed_hashes[0][1]
        # signed_hashes was empty
        return None



"""
Provides a core function for validating proofs along with a wrapper
"""

from .hashing import hash_machine
from .serializers import ValidationReceiptSerializer
import uuid
import time
import json
import os

# -------------------------------- Validation ---------------------------------


def validateProof(target_hash, proof):
    """Core validation utility

    Validates the inserted proof by comparing to the provided target hash, modifies the proof's
    status as ``True`` or ``False`` according to validation result and returns this result

    :param target_hash: the hash to be presumably attained at the end of the validation procedure (i.e.,
                        acclaimed current root-hash of the Merkle-tree having provided the proof)
    :type target_hash:  bytes
    :param proof:       the proof to be validated
    :type proof:        proof.Proof
    :returns:           validation result
    :rtype:             bool
    """

    if proof.header['generation'][:7] == 'SUCCESS':

        # Handle separately zero sublength case as always True
        if proof.header['generation'][9: -
                                      1] == 'Subtree provided by Client was empty':
            proof.header['status'] = True
            return True

        # Configure hashing parameters
        machine = hash_machine(
            hash_type=proof.header['hash_type'],
            encoding=proof.header['encoding'],
            security=proof.header['security'])

        # Perform calculation
        validated = target_hash == machine.multi_hash(
            proof.body['proof_path'], proof.body['proof_index'])

        # Inscribe new proof status according to the above calculation
        proof.header['status'] = validated

        # Print and return result
        return validated

    # generation was `FAILURE`
    proof.header['status'] = False
    return False

# ---------------------------------- Classes ----------------------------------


class ProofValidator(object):
    """Wrapper for the ``validations.validateProof`` function

    Employs the ``validations.ValidationReceipt`` class in order to organize validation results
    in an easy storable way
    """

    def __init__(self):
        pass

    def validate(self, target_hash, proof, save_dir=None):
        """Wraps ``validations.validateProof``, returning a validation receipt instead of a boolean

        If a ``save_dir`` has been specified, then the receipt is automatically stored in the given
        directory as a ``.json`` file named with the receipt's uuid

        :param target_hash: hash to be presumably attained at the end of the validation procedure (i.e.,
                            acclaimed top-hash of the Merkle-tree having provided the proof)
        :type target_hash:  bytes
        :param proof:       the proof to be validated
        :type save_dir:     [optional] Relative path with respect to the current working directory of the
                            directory where to save the generated receipt. If specified, the generated
                            receipt will be saved within this directory as a ``.json`` file named with
                            the receipt's uuid. Otherwise, then generated receipt will *not* be
                            automatically stored in any file.
        :param save_dir:    str
        :type proof:        proof.Proof
        :rtype:             validations.ValidationReceipt
        """
        validated = validateProof(target_hash=target_hash, proof=proof)

        receipt = ValidationReceipt(
            proof_uuid=proof.header['uuid'],
            proof_provider=proof.header['provider'],
            result=validated
        )

        if save_dir:
            with open(
                os.path.join(
                    save_dir,
                    '{}.json'.format(receipt.header['uuid'])
                ),
                'w'
            ) as output_file:
                json.dump(
                    receipt.serialize(),
                    output_file,
                    sort_keys=True,
                    indent=4)

        return receipt


class ValidationReceipt(object):
    """Encapsulates the result of proof validation

    :param proof_uuid:     uuid of the validated proof (time-based)
    :type proof_uuid:      str
    :param proof_provider: uuid of the Merkle-tree having provided the proof
    :type proof_provider:  str
    :param result:         Validation result (``True`` iff the proof was found to be valid)
    :type result:          bool

    Instead of providing the above arguments corresponding to `*args`, a ``ValidationReceipt`` object may also
    be constructed in the following ways by employing `**kwargs` in order to load the JSON string of a
    given validation-receipt ``r``:

    >>> from pymerkle.valiation_receipts import ValidationReceipt
    >>> s = ValidationReceipt(from_json=r.JSONstring())
    >>> t = ValidationReceipt(from_dict=json.loads(r.JSONstring()))

    .. note:: Constructing receipts in the above ways is a genuine *replication*, since the constructed
              receipts ``s`` and ``t`` have the same *uuid* and *timestamps* as ``r``

    :ivar header:                   (*dict*) Contains the keys *uuid*, *timestamp*, *validation_moment*
    :ivar header.uuid:              (*str*) uuid of the validation (time-based)
    :ivar header.timestamp:         (*str*) Validation moment (msecs) from the start of time
    :ivar header.validation_moment: (*str*) Validation moment in human readable form
    :ivar body:                     (*dict*) Contains the keys *proof_uuid*, *proof_provider*, *result* (see below)
    :ivar body.proof_uuid:          (*str*) See the homonymous argument of the constructor
    :ivar body.proof_provider:      (*str*) See the homonymous argument of the constructor
    :ivar body.result:              (*bool*) See the homonymous argument of the constructor
    """

    def __init__(self, *args, **kwargs):
        if args:                            # Assuming positional arguments by default
            self.header = {
                'uuid': str(uuid.uuid1()),  # Time based receipt id
                'timestamp': int(time.time()),
                'validation_moment': time.ctime(),
            }

            self.body = {
                'proof_uuid': args[0],
                'proof_provider': args[1],
                'result': args[2]
            }
        else:
            if kwargs.get('from_dict'):     # Importing receipt from dict
                self.header = kwargs.get('from_dict')['header']
                self.body = kwargs.get('from_dict')['body']
            elif kwargs.get('from_json'):   # Importing receipt form JSON text
                receipt_dict = json.loads(kwargs.get('from_json'))
                self.header = receipt_dict['header']
                self.body = receipt_dict['body']
            else:                           # Standard creation of a receipt
                self.header = {
                    'uuid': str(uuid.uuid1()),  # Time based receipt id
                    'timestamp': int(time.time()),
                    'validation_moment': time.ctime(),
                }

                self.body = {
                    'proof_uuid': kwargs.get('proof_uuid'),
                    'proof_provider': kwargs.get('proof_provider'),
                    'result': kwargs.get('result')
                }

    def __repr__(self):

        return '\n    ----------------------------- VALIDATION RECEIPT -----------------------------\
                \n\
                \n    uuid           : {uuid}\
                \n\
                \n    timestamp      : {timestamp} ({validation_moment})\
                \n\
                \n    proof-uuid     : {proof_uuid}\
                \n    proof-provider : {proof_provider}\
                \n\
                \n    result         : {result}\
                \n\
                \n    ------------------------------- END OF RECEIPT -------------------------------\
                \n'.format(
            uuid=self.header['uuid'],
            timestamp=self.header['timestamp'],
            validation_moment=self.header['validation_moment'],
            proof_uuid=self.body['proof_uuid'],
            proof_provider=self.body['proof_provider'],
            result='VALID' if self.body['result'] else 'NON VALID')

# ------------------------------- Serialization --------------------------

    def serialize(self):
        """ Returns a JSON entity with the receipt's attributes as key-value pairs

        :rtype: dict
        """
        serializer = ValidationReceiptSerializer()
        return serializer.default(self)

    def JSONstring(self):
        """Returns a nicely stringified version of the receipt's JSON serialized form

        .. note:: The output of this function is to be passed into the ``print`` function

        :rtype: str
        """
        return json.dumps(
            self,
            cls=ValidationReceiptSerializer,
            sort_keys=True,
            indent=4)



"""
Provides the base class for the Merkle-tree's nodes and an inheriting class for its leaves
"""

from .serializers import NodeSerializer
import json

# Prefices to be used for nice tree printing
L_BRACKET_SHORT = '\u2514' + '\u2500'           # └─
L_BRACKET_LONG = '\u2514' + 2 * '\u2500'        # └──
T_BRACKET = '\u251C' + 2 * '\u2500'             # ├──
VERTICAL_BAR = '\u2502'                         # │


class Node(object):
    """Base class for the nodes of a Merkle-tree

    :param hash_function: hash function to be used for encryption. Should be the ``.hash``
                          method of the containing Merkle-tree
    :type hash_function:  method
    :param encoding:      Encoding type to be used when decoding the hash stored by the node.
                          Should coincide with the containing Merkle-tree's encoding type.
    :type encoding:       str
    :param record:        [optional] the record to be encrypted within the node. If provided,
                          then the node is considered to be a leaf and ``stored_hash`` should
                          *not* be provided.
    :type record:         str or bytes or bytearray
    :param left:          [optional] the node's left parent. If not provided, then the node
                          is considered to be a leaf
    :type left:           nodes.Node
    :param right:         [optional] the node's right parent. If not provided, then the node
                          is considered to be a leaf
    :type right:          nodes.Node
    :param stored_hash:   [optional] The hash to be stored at creation by the node (after encoding).
                          If provided, then ``record`` should *not* be provided.
    :type stored_hash:    str

    .. warning:: *Either* ``record`` *or* ``stored_hash`` should be provided.

    :ivar stored_hash:   (*bytes*) The hash currently stored by the node
    :ivar left:          (*nodes.Node*) The node's left parent. Defaults to ``None`` if the node is a leaf
    :ivar right:         (*nodes.Node*) The node's right parent. Defaults to ``None`` if the node is a leaf
    :ivar child:         (*nodes.Node*) The node's child parent. Defaults to ``None`` if the node is a root
    :ivar encoding:      (*str*) The node's encoding type. Used for decoding its stored hash when printing
    """

    def __init__(
            self,
            hash_function,
            encoding,
            record=None,
            left=None,
            right=None,
            stored_hash=None):
        self.left, self.right, self.child = None, None, None

        # Stored for decoding when printing
        self.encoding = encoding

        if left is None and right is None:  # leaf case (parentless node)
            self.stored_hash = hash_function(record) if stored_hash is None\
                else bytes(stored_hash, encoding)
        # Interior case (node with exactly two parents)
        elif record is None:
            left.child, right.child = self, self
            self.left, self.right = left, right
            self.stored_hash = hash_function(
                left.stored_hash, right.stored_hash)

# ------------------------- Representation formatting --------------------

    def __repr__(self):
        """Overrides the default implementation

        Sole purpose of this function is to easy print info about a node by just invoking it at console

        .. warning:: Contrary to convention, the output of this implementation is *not* insertible to the ``eval`` function
        """
        def memory_id(obj): return str(
            hex(id(obj))) if obj else '{} ({})'.format(None, hex(id(obj)))

        return '\n    memory-id    : {memory_id}\
                \n    left parent  : {left}\
                \n    right parent : {right}\
                \n    child        : {child}\
                \n    hash         : {hash}\n'\
                .format(memory_id=memory_id(self),
                        left=memory_id(self.left),
                        right=memory_id(self.right),
                        child=memory_id(self.child),
                        hash=self.stored_hash.decode(self.encoding))

    def __str__(self, encoding=None, level=0, indent=3, ignore=[]):
        """Overrides the default implementation. Designed so that inserting the node as an argument to ``print``
        displays the subtree having that node as root.

        Sole purpose of this function is to be used for printing Merkle-trees in a terminal friendly way,
        similar to what is printed at console when running the ``tree`` command of Unix based platforms.

        :param encoding: [optional] encoding type to be used for decoding the node's current stored hash
        :type encoding:  str
        :param level:    [optional] Defaults to ``0``. Should be always left equal to the *default* value
                         when called externally by the user. Increased by one whenever the function is
                         recursively called so that track be kept of depth while printing
        :type level:     int
        :param indent:   [optional] the horizontal depth at which each level of the tree will be indented with
                         respect to the previous one; increase it to achieve better visibility of the tree's structure.
                         Defaults to 3.
        :type indent:    int
        :param ignore:   [optional] Defaults to the empty list ``[]``. Should be always left equal to the *default* value
                         when called externally by the user. Augmented appropriately whenever the function is recursively
                         called so that track be kept of the positions where vertical bars should be omitted
        :type ignore:    list of integers
        :rtype:          str

        .. note:: The left parent of each node is printed *above* the right one
        """
        if level == 0:
            output = '\n'
            if not self.isLeftParent() and not self.isRightParent():  # root case
                output += ' ' + L_BRACKET_SHORT
        else:
            output = (indent + 1) * ' '

        for i in range(1, level):
            if i not in ignore:
                output += ' ' + VERTICAL_BAR  # Include vertical bar
            else:
                output += 2 * ' '
            output += indent * ' '

        new_ignore = ignore[:]
        del ignore

        if self.isLeftParent():
            output += ' ' + T_BRACKET
        if self.isRightParent():
            output += ' ' + L_BRACKET_LONG
            new_ignore.append(level)

        encoding = encoding if encoding else self.encoding
        output += self.stored_hash.decode(encoding=encoding) + '\n'
        if not isinstance(self, Leaf):  # Recursive step
            output += self.left.__str__(
                encoding=encoding,
                level=level + 1,
                indent=indent,
                ignore=new_ignore)
            output += self.right.__str__(
                level=level + 1,
                encoding=encoding,
                indent=indent,
                ignore=new_ignore)
        return output

# ----------------------------- Boolean functions ------------------------

    def isLeftParent(self):
        """Checks if the node is a left parent

        :returns: ``True`` iff the node is the ``.left`` attribute of some other
                  node inside the containing Merkle-tree
        :rtype:   bool
        """
        if self.child is not None:
            return self == self.child.left
        return False

    def isRightParent(self):
        """Checks if the node is a right parent

        :returns: ``True`` iff the node is the ``.right`` attribute of some other
                  node inside the containing Merkle-tree
        :rtype:   bool
        """
        if self.child is not None:
            return self == self.child.right
        return False

# ------------------------- Merkle-tree updating tools -------------------

    def descendant(self, degree):
        """ Detects and returns the node that is ``degree`` steps upwards within
        the containing Merkle-tree

        .. note:: Descendant of degree ``0`` is the node itself, descendant of degree ``1``
                  is the node's child, etc.

        :param degree: depth of descendancy. Must be non-negative
        :type degree:  int
        :returns:      the descendant corresdponding to the requested depth
        :rtype:        nodes.Node

        .. note:: Returns ``None`` if the requested depth of dependancy exceeds possibilities
        """
        if degree == 0:
            descendant = self
        else:
            try:
                descendant = self.child.descendant(degree - 1)
            except AttributeError:
                descendant = None
        return descendant

    def recalculate_hash(self, hash_function):
        """Recalculates the node's hash under account of its parents' new hashes

        This method is to be invoked for all non-leaf nodes of the Merkle-tree's rightmost branch
        every time a new leaf is appended into the tree

        :param hash_function: hash function to be used during recalculation (thought of as
                              the ``.hash`` method of the containing Merkle-tree)
        :type hash_function:  method

        .. warning:: Only for interior nodes (i.e., with two parents), fails in case of leaf nodes
        """
        self.stored_hash = hash_function(
            self.left.stored_hash, self.right.stored_hash)


# ------------------------------- Serialization --------------------------


    def serialize(self):
        """ Returns a JSON entity with the node's attributes as key-value pairs

        :rtype: dict

        .. note:: The ``.child`` attribute is excluded from JSON formatting of nodes in order
                  for circular reference error to be avoided.
        """
        serializer = NodeSerializer()
        return serializer.default(self)

    def JSONstring(self):
        """Returns a nicely stringified version of the node's JSON serialized form

        .. note:: The output of this function is to be passed into the ``print`` function

        :rtype: str
        """
        return json.dumps(self, cls=NodeSerializer, sort_keys=True, indent=4)

# -------------------------------- End of class --------------------------


class Leaf(Node):
    """Class for the leafs of Merkle-tree (parentless nodes)

    :param hash_function: hash function to be used for encryption (only once). Should be the ``.hash``
                          attribute of the containing Merkle-tree
    :type hash_function:  method
    :param encoding:      Encoding type to be used when decoding the hash stored by the node.
                          Should coincide with the containing Merkle-tree's encoding type.
    :type encoding:       str
    :param record:        [optional] The record to be encrypted within the leaf. If provided, then
                          ``stored_hash`` should *not* be provided.
    :type record:         str or bytes or bytearray
    :param stored_hash:   [optional] The hash to be stored at creation by the leaf (after encoding).
                          If provided, then ``record`` should *not* be provided.
    :type stored_hash:    str

    .. warning:: *Either* ``record`` *or* ``stored_hash`` should be provided.
    """

    def __init__(self, hash_function, encoding, record=None, stored_hash=None):
        Node.__init__(
            self,
            record=record,
            left=None,
            right=None,
            hash_function=hash_function,
            encoding=encoding,
            stored_hash=stored_hash)



import pytest
from pymerkle import MerkleTree
from pymerkle.hashing import hash_machine


def test_tree_constructor_with_records():
    tree_1 = MerkleTree(*(bytes('{}-th record'.format(i), 'utf-8')
                          for i in range(0, 1000)))
    tree_2 = MerkleTree()
    for i in range(1000):
        tree_2.update('{}-th record'.format(i))
    assert tree_1.rootHash() == tree_2.rootHash()


# Construct standard Merkle-tree and hash function
tree = MerkleTree()
machine = hash_machine()
hash = machine.hash

# Transactions the trees will gradually be updated with
t_1, t_2, t_3, t_4, t_5, t_6, t_7, t_8, t_9, t_10, t_11 = \
    'ingi', 'rum', 'imus', 'noc', 'te', 'et', 'con', 'su', 'mi', 'mur', 'igni'


def test_0_leaves():
    assert tree.rootHash() is None


def test_1_leaves():
    tree.update(t_1)
    assert tree.rootHash() == hash(t_1)


def test_2_leaves():
    tree.update(t_2)
    assert tree.rootHash() == hash(
        hash(t_1),
        hash(t_2)
    )


def test_3_leaves():
    tree.update(t_3)
    assert tree.rootHash() == hash(
        hash(
            hash(t_1),
            hash(t_2)
        ),
        hash(t_3)
    )


def test_4_leaves():
    tree.update(t_4)
    assert tree.rootHash() == hash(
        hash(
            hash(t_1),
            hash(t_2)
        ),
        hash(
            hash(t_3),
            hash(t_4)
        )
    )


def test_5_leaves():
    tree.update(t_5)
    assert tree.rootHash() == hash(
        hash(
            hash(
                hash(t_1),
                hash(t_2)
            ),
            hash(
                hash(t_3),
                hash(t_4)
            )
        ),
        hash(t_5)
    )


def test_7_leaves():
    tree.update(t_6)
    tree.update(t_7)
    assert tree.rootHash() == hash(
        hash(
            hash(
                hash(t_1),
                hash(t_2)
            ),
            hash(
                hash(t_3),
                hash(t_4)
            )
        ),
        hash(
            hash(
                hash(t_5),
                hash(t_6)
            ),
            hash(t_7)
        )
    )


def test_11_leaves():
    tree.update(t_8)
    tree.update(t_9)
    tree.update(t_10)
    tree.update(t_11)
    assert tree.rootHash() == hash(
        hash(
            hash(
                hash(
                    hash(t_1),
                    hash(t_2)
                ),
                hash(
                    hash(t_3),
                    hash(t_4)
                )
            ),
            hash(
                hash(
                    hash(t_5),
                    hash(t_6)
                ),
                hash(
                    hash(t_7),
                    hash(t_8)
                )
            )
        ),
        hash(
            hash(
                hash(t_9),
                hash(t_10)
            ),
            hash(t_11)
        )
    )



"""
Provides the main class for Merkle-tree objects and related functionalites
"""
from .hashing import hash_machine
from .utils import log_2, decompose
from .nodes import Node, Leaf
from .proof import Proof
from .serializers import MerkleTreeSerializer
import json
import uuid
import os
import mmap
from tqdm import tqdm

# -------------------------------- Main class ----------------------------


class MerkleTree(object):
    """Class for Merkle-trees

    :param \*records:  [optional] The records initially stored by the Merkle-tree; usually empty at construction. If
                       If provided, the tree is constructed with as many leafs from the beginning, storing the hashes
                       of the inserted records in the respective order.
    :type \*records:   str or bytes or bytearray
    :param hash_type:  [optional] Defaults to ``'sha256'``. Should be included in ``hashing.HASH_TYPES`` (upper-
                       or mixed-case with '-' instead of '_' allowed), otherwise an exception is thrown.
    :type hash_type:   str
    :param encoding:   [optional] Defaults to ``'utf_8'``. Should be included in ``hashing.ENCODINGS`` (upper-
                       or mixed-case with '-' instead of '_' allowed), otherwise an exception is thrown.
    :type encoding:    str
    :param security:   [optional] If ``False``, it deactivates defense against second-preimage attack. Defaults to ``True``.
    :type security:    bool

    :ivar uuid:       (*str*) uuid of the Merkle-tree (time-based)
    :ivar hash_type:  (*str*) See the constructor's homonymous argument
    :ivar encoding:   (*str*) See the constructor's homonymous argument
    :ivar security:   (*bool*) Iff ``True``, security measures against second-preimage attack are activated
    :ivar hash:       (*method*) Core hash functionality of the Merkle-tree
    :ivar multi_hash: (*method*) Hash functionality used by the Merkle-tree for performing inclusion tests
                      (explicitly or implicitly upon a request for consistency proof)
    """

    def __init__(
            self,
            *records,
            hash_type='sha256',
            encoding='utf-8',
            security=True):

        self.uuid = str(uuid.uuid1())

        # Hash type, encoding type and security mode configuration
        machine = hash_machine(
            hash_type=hash_type,
            encoding=encoding,
            security=security)

        # Export hash and encoding type configuration
        self.hash_type = hash_type.lower().replace('-', '_')
        self.encoding = encoding.lower().replace('-', '_')
        self.security = security
        self.hash = machine.hash
        self.multi_hash = machine.multi_hash
        del machine

        # Initialized here so that consistency-proof works in some edge cases
        self.leaves, self.nodes = [], set()

        # nodes generation
        for record in records:
            self.update(record)

# --------------------------- Boolean implementation ---------------------

    def __bool__(self):
        """Overrides the default implementation

        :returns: ``False`` iff the Merkle-tree has no nodes
        :rtype:   bool
        """
        return bool(self.nodes)

# ------------------------------------ Root ------------------------------

    def rootHash(self):
        """Returns the current root-hash of the Merkle-tree, i.e., the hash stored by its current root

        :returns: the tree's current root-hash
        :rtype:   bytes

        .. note:: Returns ``None`` if the Merkle-tree is currently empty
        """
        if self:
            return self.root.stored_hash
        return None

# ------------------------------- Representation -------------------------

    def __repr__(self):
        """Overrides the default implementation.

        Sole purpose of this function is to easy print info about the Merkle-treee by just invoking it at console.

        .. warning:: Contrary to convention, the output of this implementation is *not* insertible to the ``eval`` function
        """

        return '\n    uuid      : {uuid}\
                \n\
                \n    hash-type : {hash_type}\
                \n    encoding  : {encoding}\
                \n    security  : {security}\
                \n\
                \n    root-hash : {root_hash}\
                \n\
                \n    size      : {size}\
                \n    length    : {length}\
                \n    height    : {height}\n' .format(
            uuid=self.uuid,
            hash_type=self.hash_type.upper().replace('_', '-'),
            encoding=self.encoding.upper().replace('_', '-'),
            security='ACTIVATED' if self.security else 'DEACTIVATED',
            root_hash=self.rootHash().decode(
                encoding=self.encoding) if self else '',
            size=len(self.nodes),
            length=len(self.leaves),
            height=self.height())

    def length(self):
        """Returns the Merkle-tree's current length, i.e., the number of its leaves

        :rtype: int
        """
        return len(self.leaves)

    def size(self):
        """Returns the current number of the Merkle-tree's nodes

        :rtype: int
        """
        return len(self.nodes)

    def height(self):
        """Calculates and returns the Merkle-tree's current height

        .. note:: Since the tree is by construction binary *balanced*, its height coincides
                  with the length of its leftmost branch

        :rtype: int
        """
        length = len(self.leaves)
        if length > 0:
            return log_2(length) + 1\
                if length != 2**log_2(length) else log_2(length)
        return 0

    def __str__(self, indent=3):
        """Overrides the default implementation.

        Designed so that inserting the Merkle-tree as an argument to ``print`` displays it in a terminal
        friendly way. In particular, printing the tree is similar to what is printed at console when running
        the ``tree`` command of Unix based platforms.

        :param indent: [optional] The horizontal depth at which each level will be indented with respect to
                       its previous one. Defaults to ``3``.
        :type indent:  int
        :rtype:        str

        .. note:: The left parent of each node is printed *above* the right one
        """
        if self:
            return self.root.__str__(indent=indent, encoding=self.encoding)
        return ''

    def display(self, indent=3):
        """Prints the Merkle-tree in a terminal friendy way

        Printing the tree is similar to what is printed at console when running the ``tree`` command of
        Unix based platforms

        :param indent: [optional] the horizontal depth at which each level will be indented with respect to
                       its previous one. Defaults to ``3``.
        :type indent:  int

        .. note:: The left parent of each node is printed *above* the right one
        """
        print(self.__str__(indent=indent))

# --------------------------------- Comparison ---------------------------

    def __eq__(self, other):
        """Implements the ``==`` operator

        :param other: the Merkle-tree to compare with
        :type other:  tree.MerkleTree

        Since trees with the same number of leaves have always identical structure, equality
        between Merkle-trees is established by just comparing their current root-hashes.
        """
        return isinstance(other, MerkleTree) and \
            self.rootHash() == other.rootHash()

    def __ge__(self, other):
        """Implements the ``>=`` operator

        :param other: the Merkle-tree to compare with
        :type other:  tree.MerkleTree

        A Merkle-tree is greater than or equal to another Merkle-tree iff the latter may be detected
        inside the former as a previous state of it.
        """
        return isinstance(
            other,
            MerkleTree) and self.inclusionTest(
            old_hash=other.rootHash(),
            sublength=other.length())

    def __gt__(self, other):
        """Implements the ``>`` operator

        :param other: the Merkle-tree to compare with
        :type other:  tree.MerkleTree

        A Merkle-tree is strictly greater than another Merkle-ree iff the latter may be detected inside
        the former as a previous state of it *and* their current root-hashes do not coincide (or,
        equivalently, their current lengths do not coincide).
        """
        return self >= other and self.rootHash() != other.rootHash()

# ---------------------------------- Updating ----------------------------

    def update(self, record=None, stored_hash=None):
        """Updates the Merkle-tree by storing the hash of the inserted record in a newly-created leaf,
        restructeres the tree appropriately and recalculates all necessary interior hashes

        :param record:      [optional] The record whose hash is to be stored into a new leaf.
                            If provided, then ``stored_hash`` should *not* be provided.
        :type record:       str or bytes or bytearray
        :param stored_hash: [optional] The hash to be stored by the new leaf (after encoding).
                            If provided, then ``record`` should *not* be provided.
        :type stored_hash:  str

        .. warning:: *Either* ``record`` *or* ``stored_hash`` should be provided.
        """
        if self:

            # Height of *full* binary subtree with maximum
            # possible length containing the rightmost leaf
            last_power = decompose(len(self.leaves))[-1]

            # Detect root of the above rightmost *full* binary subtree
            last_subroot = self.leaves[-1].descendant(degree=last_power)

            # Store new record to new leaf
            new_leaf = Leaf(
                hash_function=self.hash,
                encoding=self.encoding,
                record=record,
                stored_hash=stored_hash)

            # Assimilate new leaf
            self.leaves.append(new_leaf)
            self.nodes.add(new_leaf)

            # Save child info before bifurcation
            old_child = last_subroot.child

            # Create bifurcation node
            new_child = Node(
                record=None,
                left=last_subroot,
                right=new_leaf,
                hash_function=self.hash,
                encoding=self.encoding)
            self.nodes.add(new_child)

            # Bifurcate
            if not old_child:  # last_subroot was previously root

                self.root = new_child

            else:  # last_subroot was previously right parent

                # Interject bifurcation node
                old_child.right = new_child
                new_child.child = old_child

                # Recalculate hashes only at the rightmost branch of the tree
                current_node = old_child
                while current_node:
                    current_node.recalculate_hash(self.hash)
                    current_node = current_node.child

        else:  # void case
            new_leaf = Leaf(
                hash_function=self.hash,
                encoding=self.encoding,
                record=record,
                stored_hash=stored_hash)
            self.leaves, self.nodes, self.root = [
                new_leaf], set([new_leaf]), new_leaf

# ------------------------ Export to and load from file ------------------

    def export(self, file_path):
        """Exports the minimum required information into the provided file, so that the Merkle-tree can be
        reloaded in its current state from that file.

        The final file will contain a JSON entity with keys ``header`` (containing the parameters ``hash_type``,
        ``encoding`` and ``security`` of the tree) and ``hashes``, mapping to the digests currently stored by
        the tree's leaves in respective order.

        .. note:: Reconstruction of the tree is (cf. the ``loadFromFile`` static method) is uniquely determined
                  by the sequence of ``hashes`` due to the specific properties of the ``.update`` method.

        :param file_path: relative path of the file to export to with respect to the current
                          working directory
        :type file_path:  str
        """
        with open(file_path, 'w') as f:
            json.dump(self.serialize(), f, indent=4)

    @staticmethod
    def loadFromFile(file_path):
        """Loads a Merkle-tree from the provided file, the latter being the result of an export
        (cf. the ``.export`` method).

        :param file_path: relative path of the file to load from with respect to the current
                          working directory
        :type file_path:  str
        :returns:         the Merkle-tree laoded from the provided file
        :rtype:           tree.MerkleTree

        .. warning:: Raises ``KeyError`` if the provided file is not as prescribed (cf. the ``.export`` method)
        .. note :: Raises ``JSONDecodeError`` if the provided file could not be deserialized
        .. note :: Raises ``FileNotFoundError`` if the provided file does not exist
        """
        try:
            with open(file_path, 'r') as f:
                loaded_object = json.load(f)
        except (FileNotFoundError, JSONDecodeError):
            raise
        try:
            loaded_tree = MerkleTree(
                hash_type=loaded_object['header']['hash_type'],
                encoding=loaded_object['header']['encoding'],
                security=loaded_object['header']['security'])
        except KeyError:
            raise
        tqdm.write('\nFile has been loaded')
        for hash in tqdm(loaded_object['hashes'], desc='Retreiving tree...'):
            loaded_tree.update(stored_hash=hash)
        tqdm.write('Tree has been retreived')
        return loaded_tree

# --------------------------------- Encryption ---------------------------

    def encryptRecord(self, record):
        """Updates the Merkle-tree by storing the hash of the inserted record in a newly-created leaf,
        restructeres the tree appropriately and recalculates all necessary interior hashes

        :param record: the record whose hash is to be stored into a new leaf
        :type record:  str or bytes or bytearray
        """
        self.update(record=record)

    def encryptFileContent(self, file_path):
        """Encrypts the provided file as a single new leaf into the Merkle-tree

        More accurately, it updates the Merkle-tree with *one* newly created leaf storing
        the digest of the provided file's content (cf. doc of the ``.update`` method).

        :param file_path: relative path of the file under encryption with respect to the
                          current working directory
        :type file_path:  str

        .. note:: Raises ``FileNotFoundError`` if the specified file does not exist
        """
        try:
            with open(os.path.abspath(file_path), 'rb') as f:
                # ~ NOTE: File should be opened in binary mode so that its content remains
                # ~ bytes and no decoding is thus needed during hashing (otherwise byte
                # ~ 0x80 would for example be unreadable by 'utf-8' codec)
                content = f.read()  # bytes
        except FileNotFoundError:
            raise
        self.update(record=content)

    def encryptFilePerLog(self, file_path):
        """Encrypts per log the data of the provided file into the Merkle-tree

        More accurately, it successively updates the Merkle-tree with each line of the
        provided file (cf. doc of the ``.update`` method) in the respective order

        :param file_path: relative path of the file under enryption with respect to
                          the current working directory
        :type file_path:  str

        .. note:: Raises ``FileNotFoundError`` if the specified file does not exist
        """
        absolute_file_path = os.path.abspath(file_path)

        # ~ tqdm needs to know in advance the total number of
        # ~ lines so that it can display the progress bar
        number_of_lines = 0
        try:
            with open(absolute_file_path, 'r+') as file:
                # Use memory-mapped file support to count lines
                buffer = mmap.mmap(file.fileno(), 0)
        except FileNotFoundError:
            raise
        else:
            while buffer.readline():
                number_of_lines += 1

        tqdm.write('')
        # Start line by line encryption
        for line in tqdm(open(absolute_file_path, 'rb'),
                         # ~ NOTE: File should be opened in binary mode so that its content remains
                         # ~ bytes and no decoding is thus needed during hashing (otherwise byte
                         # ~ 0x80 would for example be unreadable by 'utf-8' codec)
                         desc='Encrypting log file',
                         total=number_of_lines):
            self.update(record=line)
        tqdm.write('Encryption complete\n')

    def encryptObject(self, object, sort_keys=False, indent=0):
        """Encrypts the provided object as a single new leaf into the Merkle-tree

        More accurately, it updates the Merkle-tree with *one* newly created leaf storing the digest
        of the provided object's stringified version (cf. doc of the ``.update`` method).

        :param object:    the JSON entity under encryption
        :type objec:      dict
        :param sort_keys: [optional] If ``True``, then the object's keys are alphabetically stored
                          before its stringification. Defaults to ``False``.
        :type sort_keys:  bool
        :param indent:    [optional] Specifies key indentaion upon stringification of the
                          provided object. Defaults to ``0``.
        :type indent:     int
        """
        self.update(
            record=json.dumps(
                object,
                sort_keys=sort_keys,
                indent=indent))

    def encryptObjectFromFile(self, file_path, sort_keys=False, indent=0):
        """Encrypts the object within the provided ``.json`` file as a single new leaf into the Merkle-tree

        More accurately, the Merkle-tree will be updated with *one* newly created leaf storing
        the digest of the stringified version of the object within the provided file.

        :param file_path: relative path of a ``.json`` file with respect to the current working directory,
                          containing *one* JSON entity.
        :type file_path:  str
        :param sort_keys: [optional] If ``True``, then the object's keys are alphabetically stored
                          before its stringification. Defaults to ``False``.
        :type sort_keys:  bool
        :param indent:    [optional] Specifies key indentaion upon stringification of the
                          provided object. Defaults to ``0``.
        :type indent:     int

        .. warning:: Raises ``JSONDecodeError`` if the provided file is not as prescribed
        .. note:: Raises ``FileNotFoundError`` if the specified file does not exist
        """
        try:
            with open(os.path.abspath(file_path), 'r') as f:
                object = json.load(f)
        except (FileNotFoundError, JSONDecodeError):
            raise
        self.encryptObject(object=object, sort_keys=sort_keys, indent=indent)

    def encryptFilePerObject(self, file_path, sort_keys=False, indent=0):
        """Encrypts per object the data of the provided ``.json`` file into the Merkle-tree

        More accurately, it successively updates the Merkle-tree with each newly created leaf storing the
        digest of the respective JSON entity from the list within the provided file.

        :param file_path: relative path of a ``.json`` file with respect to the current working directory,
                          containing a *list* of JSON entities.
        :type file_path:  str
        :param sort_keys: [optional] If ``True``, then the all objects' keys are alphabetically stored
                          before stringification. Defaults to ``False``.
        :type sort_keys:  bool
        :param indent:    [optional] Specifies uniform key indentaion upon stringification of objects.
                          Defaults to ``0``.
        :type indent:     int

        .. warning:: Raises ``ValueError`` if the provided file's content is not as prescribed
        .. note:: Raises ``JSONDecodeError`` if the provided file cannot be deserialized
        .. note:: Raises ``FileNotFoundError`` if the specified file does not exist
        """
        try:
            with open(os.path.abspath(file_path), 'r') as f:
                list_of_objects = json.load(f)
        except (FileNotFoundError, JSONDecodeError):
            raise
        if not isinstance(list_of_objects, list):
            raise ValueError('oops..')
        for object in list_of_objects:
            self.encryptObject(
                object=object,
                sort_keys=sort_keys,
                indent=indent)


# ------------------------------ Proof generation ------------------------


    def auditProof(self, arg):
        """Response of the Merkle-tree to the request of providing an audit-proof based upon
        the given argument

        :param arg: the record (if type is *str* or *bytes* or *bytearray*) or index of leaf (if type
                    is *int*) where the proof calculation must be based upon (provided from Client's Side)
        :type arg:  str or bytes or bytearray or int
        :returns:   Audit proof appropriately formatted along with its validation parameters (so that it
                    can be passed in as the second argument to the ``validations.validateProof`` function)
        :rtype:     proof.Proof

        .. warning:: Raises ``TypeError`` if the argument's type is not as prescribed
        """

        if type(arg) in (str, bytes, bytearray):
            # ~ Find the index of the first leaf having recorded the inserted argument;
            # ~ if no such leaf exists (i.e., the inserted argument has not been
            # ~ recorded into the tree), set index equal to -1 so that
            # ~ no genuine path be generated
            arg_hash = self.hash(arg)
            index = -1
            leaf_hashes = (leaf.stored_hash for leaf in self.leaves)
            count = 0
            for hash in leaf_hashes:
                if hash == arg_hash:
                    index = count
                    break
                count += 1
        elif isinstance(arg, int):
            index = arg  # Inserted type was integer
        else:
            raise TypeError

        # Calculate proof path
        proof_index, audit_path = self.audit_path(index=index)

        # Return proof nice formatted along with validation parameters
        if proof_index is not None:
            return Proof(
                generation='SUCCESS',
                provider=self.uuid,
                hash_type=self.hash_type,
                encoding=self.encoding,
                security=self.security,
                proof_index=proof_index,
                proof_path=audit_path)

        # Handles indexError case (`arg` provided by Client was not among
        # possibilities)
        failure_message = 'Index provided by Client was out of range'
        return Proof(
            generation='FAILURE ({})'.format(failure_message),
            provider=self.uuid,
            hash_type=self.hash_type,
            encoding=self.encoding,
            security=self.security,
            proof_index=None,
            proof_path=None)

    def consistencyProof(self, old_hash, sublength):
        """Response of the Merkle-tree to the request of providing a consistency-proof for the
        given parameters

        Arguments of this function amount to a presumed previous state of the Merkle-tree (root-hash
        and length respectively) provided from Client's Side

        :param old_hash:  root-hash of a presumably valid previous state of the Merkle-tree
        :type old_hash:   bytes
        :param sublength: presumable length (number of leaves) for the above previous state of the Merkle-tree
        :type sublength:  int
        :returns:         Consistency proof appropriately formatted along with its validation parameters (so that it
                          can be passed in as the second argument to the ``validations.validateProof`` function)
        :rtype:           proof.Proof

        .. note:: During proof generation, an inclusion-test is performed for the presumed previous state
                  of the Merke-tree corresponding to the provided parameters (If that test fails,
                  then the returned proof is predestined to be found invalid upon validation).
                  This is done implicitly and not by calling the ``.inclusionTest`` method
                  (whose implementation differs in that no full path of signed hashes,
                  as generated here by the ``.consistency_path`` method, needs be taken into account.)

        .. note:: Type of ``old_hash`` will be ``None`` iff the presumed previous state happens
                  be the empty one

        .. warning:: Raises ``TypeError`` if any of the arguments' type is not as prescribed
        """

        if type(old_hash) not in (bytes, type(None)) \
                or not isinstance(sublength, int):
            raise TypeError

        # Calculate proof path
        consistency_path = self.consistency_path(sublength=sublength)

        # Return proof nice formatted along with validation parameters
        if consistency_path is not None and\
           consistency_path[0] is not -1:  # Excludes zero leaves case
            proof_index, left_path, full_path = consistency_path

            # Root hash test
            if old_hash == self.multi_hash(left_path, len(left_path) - 1):
                return Proof(
                    generation='SUCCESS',
                    provider=self.uuid,
                    hash_type=self.hash_type,
                    encoding=self.encoding,
                    security=self.security,
                    proof_index=proof_index,
                    proof_path=full_path)

            # Handles inclusion test failure
            failure_message = 'Subtree provided by Client failed to be detected'
            return Proof(
                generation='FAILURE ({})'.format(failure_message),
                provider=self.uuid,
                hash_type=self.hash_type,
                encoding=self.encoding,
                security=self.security,
                proof_index=None,
                proof_path=None)

        # Handles a. zero `sublength` separately as success, so that proof
        # validation is always True in this case (cf. the `validations.validateProof`
        # method) b. incompatibility case as failure (includes the zero leaves)
        return Proof(
            generation='SUCCESS (Subtree provided by Client was empty)'
            if sublength == 0 else 'Subtree provided by Client was incompatible',
            provider=self.uuid,
            hash_type=self.hash_type,
            encoding=self.encoding,
            security=self.security,
            proof_index=None,
            proof_path=None)

# ------------------------------ Inclusion tests ------------------------------

    def inclusionTest(self, old_hash, sublength):
        """Verifies that the parameters provided correspond to a previous state of the Merkle-tree

        :param old_hash:  root-hash of a presumably valid previous state of the Merkle-tree
        :type old_hash:   bytes
        :param sublength: presumable length (number of leaves) for the afore-mentioned previous state of the Merkle-tree
        :type sublength:  int
        :returns:         ``True`` iff an appropriate path of negatively signed hashes, generated internally for
                          the provided ``sublength``, leads indeed to the provided ``old_hash``
        :rtype:           bool

        .. warning:: Raises ``TypeError`` if any of the arguments' type is not as prescribed
        """

        if type(old_hash) not in (bytes, type(None)) \
                or not isinstance(sublength, int):
            raise TypeError

        if 0 < sublength <= len(self.leaves):

            # Generate corresponding path of negatively signed hashes
            left_roots = self.principal_subroots(sublength)
            left_path = tuple([(-1, r[1].stored_hash) for r in left_roots])

            # Perform hash-test
            return old_hash == self.multi_hash(left_path, len(left_path) - 1)

        # return False  # No path of hashes was generated
        return True if sublength == 0 else False  # No path of hashes was generated


# ------------------------------ Path generation ------------------------------


    def audit_path(self, index):
        """Computes and returns the body for the audit-proof based upon the requested index.

        Body of an audit-proof consist of an *audit-path* (a sequence of signed hashes) and a
        *proof-index* (the position within the above sequence where the validation procedure
        should start from).

        :param index: index of the leaf where the audit-proof calculation should be based upon
                      (provided from Client's Side directly or indirectly in form of a record;
                      cf. the ``.auditProof`` method)
        :type index:  int
        :returns:     a tuple of signed hashes (pairs of the form *(+1/-1, bytes)*), the sign ``+1`` or ``-1``
                      indicating pairing with the right or left neighbour during proof validation respectively,
                      along with the starting point for application of hashing during proof validation.
        :rtype:       (int, tuple)

        .. note:: If the requested index is either negative or exceeds the tree's current length
                  (``IndexError``), then the nonsensical tuple ``(None, None)`` is returned.
        """

        # ~ Handle negative index case separately like an IndexError, since certain
        # ~ negative indices might otherwise be considered as valid positions
        if index < 0:
            return None, None

        try:
            current_node = self.leaves[index]
        except IndexError:
            return None, None  # Covers also the zero leaves case
        else:
            initial_sign = +1
            if current_node.isRightParent():
                initial_sign = -1
            path = [(initial_sign, current_node.stored_hash)]
            start = 0
            while current_node.child is not None:
                if current_node.isLeftParent():
                    next_hash = current_node.child.right.stored_hash
                    if current_node.child.isLeftParent():
                        path.append((+1, next_hash))
                    else:
                        path.append((-1, next_hash))
                else:
                    next_hash = current_node.child.left.stored_hash
                    if current_node.child.isRightParent():
                        path.insert(0, (-1, next_hash))
                    else:
                        path.insert(0, (+1, next_hash))
                    start += 1
                current_node = current_node.child
            return start, tuple(path)

    def consistency_path(self, sublength):
        """Computes and returns the body for any consistency-proof based upon the requested sublength.

        :param sublength: length (number of leaves) for a presumably valid previous state of the Merkle-tree
        :type sublength:  int
        :returns:         the starting point for application of hashing during proof validation, a tuple of hashes
                          signed with ``-1`` (leftmost hashes for inclusion test to be performed from the Server's
                          Side, i.e., by the Merkle-tree itself) and a tuple of signed hashes for top-hash test to
                          be performed from the Client's Side (the sign ``+1``, resp. ``-1`` indicating pairing
                          with the right or left neigbour respectively during proof validation)
        :rtype:           (int, tuple of (-1 bytes) pairs, tuple of (+1/-1 bytes) pairs)

        .. note::  Returns ``None`` for ``sublength`` equal to ``0``
        """
        if sublength is 0:
            return None  # so that it be handled as special incompatibility case

        left_roots = self.principal_subroots(sublength)
        if left_roots is not None:
            # No incompatibility issue

            right_roots = self.minimal_complement(
                subroots=[r[1] for r in left_roots])
            all_roots = left_roots + right_roots

            # Check if left_roots or right_roots is empty
            if sublength == 0 or sublength == len(self.leaves):

                # Reset all signs to minus
                all_roots = [(-1, r[1]) for r in all_roots]

                # Will start hashing successively from the end
                proof_index = len(all_roots) - 1

            else:  # i.e., neither left_roots nor right_roots is empty
                proof_index = len(left_roots) - 1

            # Collect and return only sign and hash pairs
            left_path = [(-1, r[1].stored_hash) for r in left_roots]
            full_path = [(r[0], r[1].stored_hash) for r in all_roots]
            return proof_index, tuple(left_path), tuple(full_path)

        return None  # Incompatibility issue detected

    def minimal_complement(self, subroots):
        """Complements optimally the subroot hashes detected by ``.principal_subroots`` with all necessary
        interior hashes of the Merkle-tree, so that a full consistency-path can be generated

        :param subroots: Should be some output of the ``.principal_subroots`` method
        :type subroots:  list of nodes.Node
        :returns:        a list of signed hashes complementing optimally the hashes detected by
                         ``.principal_subroots``, so that a full consistency-path be generated
        :rtype:          list of (+1/-1, bytes) pairs
        """
        if len(subroots) != 0:
            complement = []
            while subroots[-1].child is not None:
                last_root = subroots[-1]
                if last_root is last_root.child.left:
                    if last_root.child.isRightParent():
                        complement.append((-1, last_root.child.right))
                    else:
                        complement.append((+1, last_root.child.right))
                    subroots = subroots[:-1]
                else:
                    subroots = subroots[:-2]
                subroots.append(last_root.child)
            return complement
        return self.principal_subroots(len(self.leaves))

    def principal_subroots(self, sublength):
        """Detects and returns in corresponding order the roots of the *successive*, *rightmost*, *full* binary
        subtrees of maximum (and thus decreasing) ength, whose lengths sum up to the inserted argument

        Returned nodes are prepended with a sign (``+1`` or ``-1``), carrying information used in
        consistency-proof generation after extracting hashes

        :param sublength: Should be a non-negative integer smaller than or equal to the Merkle-tree's current length
        :returns:         The (signed) roots of the detected subtrees, whose hashes
                          are to be used for the generation of consistency-proofs
        :rtype:           list of *(+1/-1, nodes.Node)*

        .. note:: Returns ``None`` if the provided ``sublength`` does not fulfill the required condition
        """

        if sublength == 0:
            return []
        elif sublength > 0:
            principal_subroots = []
            powers = decompose(sublength)
            start = 0
            i = 0
            for i in range(0, len(powers)):
                next_subroot = self.subroot(start, powers[i])
                if next_subroot is not None:  # No incompatibility issue
                    if next_subroot.child and next_subroot.child.child:
                        if next_subroot.child.isLeftParent():
                            principal_subroots.append((+1, next_subroot))
                        else:
                            principal_subroots.append((-1, next_subroot))
                    else:
                        if next_subroot.isLeftParent():
                            principal_subroots.append((+1, next_subroot))
                        else:
                            principal_subroots.append((-1, next_subroot))
                    start += 2**powers[i]
                else:
                    # Incompatibility issue detected; break loop and return
                    return None
            # Principal subroot successfully detected
            if len(principal_subroots) > 0:
                # modify last sign
                principal_subroots[-1] = (+1, principal_subroots[-1][1])
            return principal_subroots
        else:  # Negative input handled as `incompatibility`
            return None

    def subroot(self, start, height):
        """
        Returns the root of the unique *full* binary subtree of the Merkle-tree, whose leftmost leaf is located
        at the given position ``start`` and whose height is equal to the given ``height``

        :param start:  index of leaf where detection of subtree should start from
        :type start:   int
        :param height: height of candidate subtree to be detected
        :type height:  int
        :returns:      root of the detected subtree
        :rtype:        nodes.Node

        .. note:: Returns ``None`` if the requested ``start`` is out of range
        """
        subroot = None

        # Detect candidate subroot
        try:
            subroot = self.leaves[start]
            i = 0
            while i < height:
                try:
                    next_node = subroot.child
                    if next_node.left is not subroot:
                        raise AttributeError
                    else:
                        subroot = subroot.child
                except AttributeError:
                    return None
                else:
                    i += 1
        except IndexError:
            return None

        # ~ Verify existence of *full* binary subtree for the above
        # ~ detected candidate subroot
        right_parent = subroot
        i = 0
        while i < height:
            if isinstance(right_parent, Leaf):
                return None  # Subtree failed to be detected
            else:
                right_parent = right_parent.right
                i += 1

        # Subroot successfully detected
        return subroot

# ---------------------------------- Clearance ---------------------------

    def clear(self):
        """Deletes all nodes of the Merkle-tree, so that its root-hash becomes ``None``
        """
        self.leaves = []
        self.nodes = set()
        self.root = None

# ------------------------------- Serialization --------------------------

    def serialize(self):
        """ Returns a JSON entity with the Merkle-trees's current characteristics and
        hashes currently stored by its leaves.

        :rtype: dict

        .. note:: This method does *not* serialize the tree structure itself, but only the info
                  about the tree's fixed configs and current state, so that the tree can be
                  retrieved from that using the ``.update`` method
        """
        serializer = MerkleTreeSerializer()
        return serializer.default(self)

    def JSONstring(self):
        """Returns a nicely stringified version of the Merkle-tree's JSON serialized form

        .. note:: The output of this method is to be passed into the ``print()`` function

        :rtype: str
        """
        return json.dumps(
            self,
            cls=MerkleTreeSerializer,
            sort_keys=True,
            indent=4)



import pytest
from pymerkle import MerkleTree, validateProof
from pymerkle.proof import Proof
import json

# Generate proofs from a one-thousand leaves Merkle-tree
tree = MerkleTree(*(bytes('{}-th record'.format(i), 'utf-8')
                    for i in range(0, 1000)))
p = tree.auditProof(666)  # Genuine index-based proof
q = tree.auditProof(1000)  # Empty index-based proof

# ------------------- Check replicates in all possible ways -------------------

p_1 = Proof(from_json=p.JSONstring())
p_2 = Proof(from_dict=json.loads(p.JSONstring()))
q_1 = Proof(from_json=q.JSONstring())
q_2 = Proof(from_dict=json.loads(q.JSONstring()))


@pytest.mark.parametrize('replicate', (p_1, p_2))
def test_p_replicates_via_serialization(replicate):
    assert p.serialize() == replicate.serialize()


@pytest.mark.parametrize('replicate', (q_1, q_2))
def test_q_replicates_via_serialization(replicate):
    assert q.serialize() == replicate.serialize()

# ----------------------- Check record based generations -----------------


p_3 = tree.auditProof('665-th record')
p_4 = tree.auditProof(b'665-th record')


@pytest.mark.parametrize('r', (p_3, p_4))
def test_record_based_audit_proof(r):
    assert tree.multi_hash(
        p.body['proof_path'],
        p.body['proof_index']) == tree.multi_hash(
        r.body['proof_path'],
        p.body['proof_index'])



from .tree import MerkleTree
from .validations import validateProof, ProofValidator

name = "pymerkle"

__version__ = "3.0.1b"
__all__ = (
    'MerkleTree',
    'validateProof',
    'ProofValidator')



"""
Provides JSON encoders used implicitly for serialization accross this library
"""
import json


class MerkleTreeSerializer(json.JSONEncoder):
    """Used implicitly in the JSON serialization of Merkle-trees.
    Extends the built-in JSON encoder for data structures.
    """

    def default(self, obj):
        """ Overrides the built-in method of JSON encoders according to the needs of this library
        """
        try:
            hash_type, encoding, security = obj.hash_type, obj.encoding, obj.security
            leaves = obj.leaves
        except TypeError:
            return json.JSONEncoder.default(self, obj)
        else:
            return {
                'header': {
                    'hash_type': hash_type,
                    'encoding': encoding,
                    'security': security},
                'hashes': [
                    leaf.stored_hash.decode(
                        encoding=encoding) for leaf in leaves]}


class NodeSerializer(json.JSONEncoder):
    """Used implicitly in the JSON serialization of nodes.
    Extends the built-in JSON encoder for data structures.
    """

    def default(self, obj):
        """ Overrides the built-in method of JSON encoders according to the needs of this library.
        """
        try:
            left, right = obj.left, obj.right
            hash = obj.stored_hash
        except TypeError:
            return json.JSONEncoder.default(self, obj)
        else:
            if isinstance(obj, Leaf):
                return {
                    'hash': hash.decode(encoding=obj.encoding)
                }
            return {
                'left': left.serialize(),
                'right': right.serialize(),
                'hash': hash.decode(encoding=obj.encoding)
            }  # Non-leaf case


class ProofSerializer(json.JSONEncoder):
    """Used implicitly in the JSON serialization of proofs.
    Extends the built-in JSON encoder for data structures.
    """

    def default(self, obj):
        """ Overrides the built-in method of JSON encoders according to the needs of this library
        """
        try:
            uuid = obj.header['uuid']
            generation = obj.header['generation']
            timestamp = obj.header['timestamp']
            creation_moment = obj.header['creation_moment']
            provider = obj.header['provider']
            hash_type = obj.header['hash_type']
            encoding = obj.header['encoding']
            security = obj.header['security']
            proof_index = obj.body['proof_index']
            proof_path = obj.body['proof_path']
            status = obj.header['status']
        except TypeError:
            return json.JSONEncoder.default(self, obj)
        else:
            return {
                'header': {
                    'uuid': uuid,
                    'generation': generation,
                    'timestamp': timestamp,
                    'creation_moment': creation_moment,
                    'provider': provider,
                    'hash_type': hash_type,
                    'encoding': encoding,
                    'security': security,
                    'status': status
                },
                'body': {
                    'proof_index': proof_index,
                    'proof_path': [
                        [sign, hash
                            if isinstance(hash, str) else hash.decode(encoding=encoding)
                         ] for (sign, hash) in proof_path]if proof_path is not None else []
                }
            }


class ValidationReceiptSerializer(json.JSONEncoder):
    """Used implicitly in the JSON serialization of validation receipts.
    Extends the built-in JSON encoder for data structures.
    """

    def default(self, obj):
        """ Overrides the built-in method of JSON encoders according to the needs of this library
        """
        try:
            uuid = obj.header['uuid']
            timestamp = obj.header['timestamp']
            validation_moment = obj.header['validation_moment']
            proof_uuid = obj.body['proof_uuid']
            proof_provider = obj.body['proof_provider']
            result = obj.body['result']
        except TypeError:
            return json.JSONEncoder.default(self, obj)
        else:
            return {
                'header': {
                    'uuid': uuid,
                    'timestamp': timestamp,
                    'validation_moment': validation_moment
                },
                'body': {
                    'proof_uuid': proof_uuid,
                    'proof_provider': proof_provider,
                    'result': result
                }
            }



"""
Provides standalone utilities invoked across the library
"""

import math


def log_2(num):
    """Computes and returns the base *2* logarithm of the given number (i.e.,
    the greatest power of *2* equal to or smaller than ``num``)

    .. note:: Given any *balanced* binary tree, whose number of leaves equals
              the inserted one, this function returns the tree's height
              (i.e., the depth of its *left-most* branch)

    :param num: the number whose logarithm is to be computed
    :type num:  int
    :returns:   the computed logarithm
    :rtype:     int

    .. note:: By convention, it returns 0 for zero argument

    :raises ValueError: for arguments smaller than zero
    """
    return 0 if num == 0 else int(math.log(num, 2))


def decompose(num):
    """Additive decomposition in decreasing powers of 2

    Given a positive integer uniquely decomposed as

    ``2 ^ (p_m) + ... + 2 ^ (p_1),  p_m > ... > p_1 >= 0``

    then the tuple ``(p_m, ..., p_1)`` is returned

    :Example:

    >>> num = 45
    >>> num == 2**5 + 2**3 + 2**2 + 1
    True
    >>>
    >>> decompose(num)
    (5, 3, 2, 0)

    :param num: the number to be decomposed
    :type num:  int
    :returns:   powers of *2* in decreasing order
    :rtype:     tuple of integers

    .. note:: Returns the nonsensical empty tuple for arguments equal to or
              smaller than zero
    """
    powers = []
    while num > 0:
        power = log_2(num)
        num -= 2**power
        powers.append(power)
    return tuple(powers)


def stringify_path(signed_hashes, encoding):
    """Returns a nicely stringified version of the inserted sequence of signed hashes.

    The printed hashes are hexadecimals, occuring after decoding the given ones according
    to the inserted encoding type.

    .. note:: The output of this function is to be passed into the ``print`` function

    :param signed_hashes: a sequence of signed hashes
    :type signed_hashes:  tuple of (+1/-1, bytes) pairs
    :param encoding:      type to be used for decoding
    :type encoding:       str
    :rtype:               str
    """
    def order_of_magnitude(num): return 0 if num == 0 else int(math.log10(num))

    def get_with_sign(num): return str(num) if num < 0 else '+' + str(num)

    if signed_hashes is not None:
        stringified_elems = []
        for i in range(len(signed_hashes)):
            elem = signed_hashes[i]
            stringified_elems.append(
                ('\n' +
                 (7 - order_of_magnitude(i)) * ' ' +
                 '[{i}]' +
                 3 * ' ' +
                 '{sign}' +
                 2 * ' ' +
                 '{hash}'). format(
                     i=i,
                     sign=get_with_sign(elem[0]),
                     hash=elem[1].decode(encoding=encoding)
                     if not isinstance(elem[1], str) else elem[1]))
        return ''.join(elem for elem in stringified_elems)
    return ''  # input was None



import pytest
import os
import json
from pymerkle import MerkleTree, hashing, encodings

HASH_TYPES = hashing.HASH_TYPES
ENCODINGS = encodings.ENCODINGS

# Generate trees and corresoponding hash machines for all combinations of
# hash and encoding types (including both security modes for each)
trees = []
hash_machines = []
for security in (True, False):
    for hash_type in HASH_TYPES:
        for encoding in ENCODINGS:
            trees.append(MerkleTree(
                hash_type=hash_type,
                encoding=encoding,
                security=security))
            hash_machines.append(hashing.hash_machine(
                hash_type=hash_type,
                encoding=encoding,
                security=security))


@pytest.mark.parametrize(
    "tree, hash_machine", [
        (trees[i], hash_machines[i]) for i in range(
            len(trees))])
def test_encryptRecord(tree, hash_machine):
    tree.encryptRecord('some kind of record...')
    assert tree.rootHash() == hash_machine.hash(
        'some kind of record...')


large_APACHE_log_path = os.path.join(
    os.path.dirname(__file__),
    'logs/large_APACHE_log')

with open(large_APACHE_log_path, 'rb') as f:
    content = f.read()


@pytest.mark.parametrize(
    "tree, hash_machine", [
        (trees[i], hash_machines[i]) for i in range(
            len(trees))])
def test_encryptFileContent(tree, hash_machine):
    tree.clear()
    tree.encryptFileContent(large_APACHE_log_path)

    assert tree.rootHash() == hash_machine.hash(content)


short_APACHE_log_path = os.path.join(
    os.path.dirname(__file__),
    'logs/short_APACHE_log')

records = []
with open(short_APACHE_log_path, 'rb') as log_file:
    for line in log_file:
        records.append(line)


@pytest.mark.parametrize("tree", trees)
def test_encryptFilePerLog(tree):
    tree.clear()
    # Update original tree directly from file
    tree.encryptFilePerLog(short_APACHE_log_path)

    clone_tree = MerkleTree(
        hash_type=tree.hash_type,
        encoding=tree.encoding,
        security=tree.security)

    for record in records:
        clone_tree.update(record)

    assert tree.rootHash() == clone_tree.rootHash()


@pytest.mark.parametrize(
    "tree, hash_machine", [
        (trees[i], hash_machines[i]) for i in range(
            len(trees))])
def test_encryptObject(tree, hash_machine):
    tree.clear()
    tree.encryptObject(object={'a': 0, 'b': 1}, sort_keys=False, indent=0)

    assert tree.rootHash() == hash_machine.hash(
        json.dumps({'a': 0, 'b': 1}, sort_keys=False, indent=0))


object_file_path = os.path.join(
    os.path.dirname(__file__),
    'objects/sample.json')

with open(object_file_path, 'r') as f:
    object_from_file = json.load(f)


@pytest.mark.parametrize(
    "tree, hash_machine", [
        (trees[i], hash_machines[i]) for i in range(
            len(trees))])
def test_encryptObjectFromFile(tree, hash_machine):
    tree.clear()
    tree.encryptObjectFromFile(object_file_path, sort_keys=False, indent=0)

    assert tree.rootHash() == hash_machine.hash(
        json.dumps(object_from_file, sort_keys=False, indent=0))


objects_file_path = os.path.join(
    os.path.dirname(__file__),
    'objects/sample-list.json')

with open(objects_file_path, 'r') as f:
    list_of_objects = json.load(f)


@pytest.mark.parametrize(
    "tree, hash_machine", [
        (trees[i], hash_machines[i]) for i in range(
            len(trees))])
def test_encryptFilePerObject(tree, hash_machine):
    tree.clear()
    tree.encryptFilePerObject(objects_file_path, sort_keys=False, indent=0)

    clone_tree = MerkleTree(
        hash_type=tree.hash_type,
        encoding=tree.encoding,
        security=tree.security)

    for object in list_of_objects:
        clone_tree.update(record=json.dumps(object, sort_keys=False, indent=0))

    assert tree.rootHash() == clone_tree.rootHash()



import pytest
import os
from pymerkle import MerkleTree, hashing, encodings, validateProof

# --------- Test intermediate success case for all possible tree types ---

HASH_TYPES = hashing.HASH_TYPES
ENCODINGS = encodings.ENCODINGS

# Files to encrypt
short_APACHE_log = os.path.join(
    os.path.dirname(__file__),
    'logs/short_APACHE_log')
RED_HAT_LINUX_log = os.path.join(
    os.path.dirname(__file__),
    'logs/RED_HAT_LINUX_log')

# Generate trees (for all combinations of hash and encoding types)
# along with valid parameters for inclusion test
trees_and_subtrees = []
for security in (True, False):
    for hash_type in HASH_TYPES:
        for encoding in ENCODINGS:
            tree = MerkleTree(
                hash_type=hash_type,
                encoding=encoding,
                security=security)
            tree.encryptFilePerLog(short_APACHE_log)
            old_hash, sublength = tree.rootHash(), tree.length()
            tree.encryptFilePerLog(RED_HAT_LINUX_log)
            trees_and_subtrees.append((tree, old_hash, sublength))


@pytest.mark.parametrize("tree, old_hash, sublength", trees_and_subtrees)
def test_inclusion_test_with_valid_parameters(tree, old_hash, sublength):
    assert tree.inclusionTest(old_hash, sublength) is True

# -------------- Test success edge case with standard Merkle-Tree --------


tree = MerkleTree()
tree.encryptFilePerLog(short_APACHE_log)
old_hash, sublength = tree.rootHash(), tree.length()
tree.encryptFilePerLog(RED_HAT_LINUX_log)


def test_inclusion_test_edge_success_case():
    assert tree.inclusionTest(tree.rootHash(), tree.length()) is True

# ---------------- Test failure cases with standard Merkle-tree ----------


def test_inclusion_test_with_zero_sublength():
    assert tree.inclusionTest(b'anything...', 0) is True


def test_inclusion_test_with_sublength_exceeding_length():
    assert tree.inclusionTest(b'anything...', tree.length()) is False


@pytest.mark.parametrize('sublength', list(i for i in range(1, tree.length())))
def test_inclusion_test_with_invalid_old_hash(sublength):
    assert tree.inclusionTest(
        b'anything except for the hash corresponding to the provided sublength',
        sublength) is False

# -------------- Test inclusion for sublength equal to power of 2 --------

# ~ Passing the following tests indicates that the bug concerning inclusion
# ~ tests (or, implicitly, consistency proofs) for sublengths equal to
# ~ powers of 2 has been fixed (in particular, the ``multi_hash``
# ~ functionality has been modified appropriately for the case of
# ~ one-member-sequences, so that this issue does not arise)


# Initialize parametrization with the empty tree (edge case)
trees_and_later_states = [(
    MerkleTree(), MerkleTree(*[str(k) for k in range(1, j)])
) for j in range(0, 10)]

# Parametrize for the first 10 powers of 2
for power in range(0, 10):
    tree = MerkleTree(*[str(i) for i in range(1, 2**power + 1)])
    for j in range(0, 10):
        trees_and_later_states.append((
            tree,
            MerkleTree(*[str(k) for k in range(1, 2**power + 1 + j)])
        ))


@pytest.mark.parametrize('tree, later_state', trees_and_later_states)
def test_inclusion_test_with_sublength_equal_to_power_of_2(tree, later_state):
    assert later_state.inclusionTest(
        old_hash=tree.rootHash(),
        sublength=tree.length()) is True


@pytest.mark.parametrize('tree, later_state', trees_and_later_states)
def test_consistency_proof_validation_with_sublength_equal_to_power_of_2(
        tree,
        later_state):
    assert validateProof(
        target_hash=later_state.rootHash(),
        proof=later_state.consistencyProof(
            old_hash=tree.rootHash(),
            sublength=tree.length())) is True



import pytest
import os
import json
import time
from pymerkle import MerkleTree, hashing, validateProof, ProofValidator
from pymerkle.validations import ValidationReceipt

# ---------------- Check receipt replicates in all possible ways ---------

tree = MerkleTree(*(bytes('{}-th record'.format(i), 'utf-8')
                    for i in range(0, 1000)))
p = tree.auditProof(666)
v = ProofValidator()
r = v.validate(target_hash=tree.rootHash(), proof=p)
r_1 = ValidationReceipt(from_json=r.JSONstring())
r_2 = ValidationReceipt(from_dict=json.loads(r.JSONstring()))


@pytest.mark.parametrize('replicate', (r_1, r_2))
def test_r_replicates_via_serialization(replicate):
    assert r.serialize() == replicate.serialize()

# ---------------------- Validation tests parametrization ----------------


HASH_TYPES = hashing.HASH_TYPES
ENCODINGS = ['utf_7',
             'utf_8',
             'utf_16',
             'utf_16_be',
             'utf_16_le',
             'utf_32',
             'utf_32_be',
             'utf_32_le']

# Files to encrypt
short_APACHE_log = os.path.join(
    os.path.dirname(__file__),
    'logs/short_APACHE_log')
RED_HAT_LINUX_log = os.path.join(
    os.path.dirname(__file__),
    'logs/RED_HAT_LINUX_log')
large_APACHE_log = os.path.join(
    os.path.dirname(__file__),
    'logs/large_APACHE_log')

# Store first log size
with open(short_APACHE_log) as first_log_file:
    first_log_size = sum(1 for line in first_log_file)

# Store second log size
with open(RED_HAT_LINUX_log) as second_log_file:
    second_log_size = sum(1 for line in second_log_file)

# ------------------- Test validate proof for empty tree case ------------

trees = []
for encoding in ENCODINGS:
    for hash_type in HASH_TYPES:
        tree = MerkleTree(
            hash_type=hash_type,
            encoding=encoding,
            security=True)
        trees.append(tree)


@pytest.mark.parametrize('tree', trees)
def test_proof_validation_for_empty_tree(tree):
    """Tests proof-validation for proofs provided by empty trees
    """
    audit_proof = tree.auditProof(arg=0)
    consistency_proof = tree.consistencyProof(
        old_hash=tree.rootHash(), sublength=0)

    assert validateProof(
        target_hash=b'anything...',
        proof=audit_proof) is False and validateProof(
        target_hash=b'anything...',
        proof=consistency_proof) is True

# ------------------------ Test audit proof validation ------------------------


audit_proofs = []
target_hashes = []
expecteds = []

for bool_1 in (True, False):  # Controls index compatibility
    for bool_2 in (True, False):  # Controls validity of target hash
        for encoding in ENCODINGS:
            for hash_type in HASH_TYPES:
                for arg in range(first_log_size):

                    # Expected value configuration
                    expecteds.append(bool_1 and bool_2)

                    # Proof-provider configuration
                    tree = MerkleTree(
                        hash_type=hash_type,
                        encoding=encoding,
                        security=True)
                    tree.encryptFilePerLog(short_APACHE_log)

                    # Proof configuration
                    if bool_1:
                        audit_proofs.append(tree.auditProof(arg=arg))
                    else:
                        audit_proofs.append(tree.auditProof(
                            arg=first_log_size + arg))

                    # Target-hash configuration
                    if bool_2:
                        target_hashes.append(tree.rootHash())
                    else:
                        target_hashes.append(b'anything else...')


@pytest.mark.parametrize(
    'audit_proof, target_hash, expected', [
        (audit_proofs[i], target_hashes[i], expecteds[i]) for i in range(
            len(audit_proofs))])
def test_index_based_audit_proof_validation_for_non_empty_tree(
        audit_proof, target_hash, expected):
    assert validateProof(
        target_hash=target_hash,
        proof=audit_proof) is expected


small_tree = MerkleTree('0', '1', '2', '3', '4', '5', '6', '7', '8', '9')
audit_proofs = []
expecteds = []

for bool in (True, False):
    for i in range(0, 10):
        if bool:
            audit_proofs.append(small_tree.auditProof(arg=str(i)))
        else:
            audit_proofs.append(small_tree.auditProof(arg=str(10 + i)))
        expecteds.append(bool)


@pytest.mark.parametrize(
    'audit_proof, expected', [
        (audit_proofs[i], expecteds[i]) for i in range(
            len(audit_proofs))])
def test_record_based_audit_proof_validation_for_non_empty_tree(
        audit_proof, expected):
    assert validateProof(
        target_hash=small_tree.rootHash(),
        proof=audit_proof) is expected

# --------------------- Test consistency proof validation ---------------------


consistency_proofs = []
target_hashes = []
expecteds = []

for bool_1 in (
        True,
        False):  # Controls subtree detection via validity of old tree hash
    for bool_2 in (
            True, False):  # Controls subtree detection via its length
        for bool_3 in (
                True, False):  # Controls subtree compatibility
            for bool_4 in (
                    True, False):  # Controls validity of target hash
                for encoding in ENCODINGS:
                    for hash_type in HASH_TYPES:

                        # Expected value configuration
                        expecteds.append(
                            bool_1 and bool_2 and bool_3 and bool_4)

                        # Proof-provider configuration
                        tree = MerkleTree(
                            hash_type=hash_type,
                            encoding=encoding,
                            security=True)

                        # Append first log
                        tree.encryptFilePerLog(short_APACHE_log)

                        # Old-tree-hash configuration
                        if bool_1:
                            old_hash = tree.rootHash()
                        else:
                            old_hash = b'anything else...'

                        # Subtree-detection configuration
                        if bool_2 and bool_3:
                            old_tree_length = first_log_size
                        elif not bool_2 and bool_3:
                            old_tree_length = first_log_size - 1
                        else:
                            old_tree_length = second_log_size + first_log_size

                        # Update the tree by appending new log
                        tree.encryptFilePerLog(RED_HAT_LINUX_log)

                        # Generate proof for the above configurations
                        consistency_proofs.append(
                            tree.consistencyProof(
                                old_hash=old_hash,
                                sublength=old_tree_length))

                        # Target-hash configuration
                        if bool_4:
                            target_hashes.append(tree.rootHash())
                        else:
                            target_hashes.append(b'anything else...')


@pytest.mark.parametrize(
    'consistency_proof, target_hash, expected', [
        (consistency_proofs[i], target_hashes[i], expecteds[i]) for i in range(
            len(consistency_proofs))])
def test_consistency_proof_validation_for_non_empty_tree(
        consistency_proof, target_hash, expected):
    assert validateProof(
        target_hash=target_hash,
        proof=consistency_proof) is expected


# ------------------------- Test proof validator object ------------------

# Proof provider (a typical SHA256/UTF-8 Merkle-Tree with defense against
# second-preimage attack)
tree = MerkleTree()

file_dir = os.path.dirname(__file__)

# Proof validator
validator = ProofValidator()

# Clean validations directory before running the test
file_list = os.listdir(os.path.join(file_dir, 'receipts'))
for file in file_list:
    os.remove(os.path.join(file_dir, 'receipts', file))

# Feed tree with logs gradually and generate consistency proof for each step
proofs = []
target_hashes = []
for log_file in (large_APACHE_log, RED_HAT_LINUX_log, short_APACHE_log):
    old_hash = tree.rootHash()
    old_length = len(tree.leaves)
    tree.encryptFilePerLog(log_file)
    proofs.append(
        tree.consistencyProof(
            old_hash=old_hash,
            sublength=old_length))
    target_hashes.append(tree.rootHash())


@pytest.mark.parametrize(
    'proof, target_hash', [
        (proofs[i], target_hashes[i]) for i in range(
            len(proofs))])
def test_ProofValidator(proof, target_hash):
    receipt = validator.validate(
        proof=proof,
        target_hash=target_hash,
        save_dir=os.path.join(
            file_dir,
            'receipts'))
    receipt_file_path = os.path.join(
        file_dir,
        'receipts',
        '{}.json'.format(
            receipt.header['uuid']))
    with open(receipt_file_path) as receipt_file:
        receipt_clone = json.load(receipt_file)
    assert receipt.serialize() == receipt_clone



import pytest
import os
import json
from pymerkle import MerkleTree

# Clean exports dir before running the test
for file in os.listdir(os.path.join(os.path.dirname(__file__), 'exports')):
    os.remove(
        os.path.join(
            os.path.dirname(
                __file__),
            'exports',
            file))

tree = MerkleTree()
for i in range(12):
    tree.update(record='{}-th record'.format(i))
export_path = os.path.join(
    os.path.dirname(__file__),
    'exports',
    '{}.json'.format(
        tree.uuid))
tree.export(file_path=export_path)
with open(export_path, 'r') as f:
    exported_version = json.load(f)


def test_export():
    assert exported_version == {
        "header": {
            "encoding": "utf_8",
            "hash_type": "sha256",
            "security": True
        },
        "hashes": [
            "a08665f5138f40a07987234ec9821e5be05ecbf5d7792cd4155c4222618029b6",
            "3dbbc4898d7e909de7fc7bb1c0af36feba78abc802102556e4ea52c28ccb517f",
            "45c44059cf0f5a447933f57d851a6024ac78b44a41603738f563bcbf83f35d20",
            "b5db666b0b34e92c2e6c1d55ba83e98ff37d6a98dda532b125f049b43d67f802",
            "69df93cbafa946cfb27c4c65ae85222ad5c7659237124c813ed7900a7be83e81",
            "9d6761f55a3e87166d2ea6d00db9c88159c893674a8420cb8d32c35dbb791fd4",
            "e718ae6ea64cb37a593654f9c0d7ec81d11498fdd94fc5473b999cd6c00d05c6",
            "ad2c93dd91eafb31ad91deb8c1b318b126957608d13bfdba209a5f17ecf22503",
            "cdc94791cd56543e1b28b21587c76f7cb45203fa7b1b8aa219e6ccc527a0d0d9",
            "828a54ce62ae58e01271a3bde442e0fa6bfa758b2816dd39f873718dfa27634a",
            "5ebc41746c5fbcfd8d32eef74f1aaaf02d6da8ff94426855393732db8b73126a",
            "b70665abe265a88bc68ec625154746457a2ba7ecb5a7fc792e9443f618fc93fd"
        ]
    }


def test_loadFromFile():
    assert tree.rootHash() == MerkleTree.loadFromFile(
        file_path=export_path).rootHash()



""" Performs second-preimage attack against Merkle-trees of all possible hash- and encoding-types
for *both* possible security modes. Attack should succeed only when the tree's security mode is
deactiveated (i.e., *iff* its ``.security`` attribute is ``False`` by construction).

Attack Schema
-------------

               Original tree               Attacker's tree

                     A                           A
                   /   \                       /   \
                 B       C = h(FG)           B      C    ------> [injected leaf]
                / \     / \                 / \     |
 hashes:       D   E   F   G               D   E   (FG)  <------ [forged record]
               |   |   |   |               |   |
records:       d   e   f   g               d   e


Concatenate the hashes stored by the 3-rd and 4-th leaves and append the result (i.e., its hash)
as the 3-rd leaf, leaving the 1-st and 2-nd leaves untouched
"""
import pytest
from pymerkle import MerkleTree, hashing, encodings

# Generate trees for all combinations of hash and encoding types
# (including both security modes for each)
HASH_TYPES = hashing.HASH_TYPES
ENCODINGS = encodings.ENCODINGS
trees = []
for security in (True, False):
    for hash_type in HASH_TYPES:
        for encoding in ENCODINGS:
            trees.append(
                MerkleTree(
                    'a', 'b', 'c', 'd',  # original records
                    hash_type=hash_type,
                    encoding=encoding,
                    security=security
                )
            )


@pytest.mark.parametrize("original_tree", trees)
def test_defense_against_second_preimage_attack(original_tree):
    # Construct forged record
    F = original_tree.leaves[2].stored_hash.decode(
        encoding=original_tree.encoding)
    G = original_tree.leaves[3].stored_hash.decode(
        encoding=original_tree.encoding)
    forged_record = '%s%s' % (F, G)
    # Construct attacker's tree
    attacker_tree = MerkleTree(
        'a', 'b', forged_record,
        hash_type=original_tree.hash_type,
        encoding=original_tree.encoding,
        security=original_tree.security
    )
    # Check if the attacker has found the original root-hash
    if original_tree.security:
        assert original_tree.rootHash() != attacker_tree.rootHash()
    else:
        assert original_tree.rootHash() == attacker_tree.rootHash()



import pytest
from pymerkle import hashing, encodings

import hashlib

HASH_TYPES = tuple(hashing.HASH_TYPES)
excluded_ENCODINGS = (
    'utf_16',
    'utf_16_be',
    'utf_16_le',
    'utf_32',
    'utf_32_be',
    'utf_32_le')
ENCODINGS = (e for e in encodings.ENCODINGS if e not in excluded_ENCODINGS)

# Hard-coded string to be used for testing
message = 'oculusnonviditnecaurisaudivit'

# Generate hash machines for any combination of hash and encoding types
# (including both security modes for each )
hash_machines = []
hash_types = []
encodings = []
securities = []
bytes_messages = []
bytearray_messages = []
for security in {True, False}:
    for hash_type in HASH_TYPES:
        for encoding in ENCODINGS:
            hash_machines.append(hashing.hash_machine(
                hash_type=hash_type,
                encoding=encoding,
                security=security))
            hash_types.append(hash_type)
            encodings.append(encoding)
            securities.append(security)
            bytes_messages.append(bytes(message, encoding=encoding))
            bytearray_messages.append(bytearray(message, encoding=encoding))


@pytest.mark.parametrize(
    "hash_machine, hash_type, encoding, security", [
        (hash_machines[i], hash_types[i], encodings[i], securities[i]) for i in range(
            len(hash_machines))])
def test_string_hash(hash_machine, hash_type, encoding, security):
    """Tests single string hashing for all combinations of hash and encoding types
    """
    if hash_type == 'sha256' and encoding == 'utf_8' and security:  # Genuinely activated security standards
        assert hash_machine.hash(message) == bytes(
            getattr(hashlib, hash_type)(
                ('\x00' + message).encode(encoding=encoding)
            ).hexdigest(), encoding=encoding)
    else:
        assert hash_machine.hash(message) == bytes(
            getattr(hashlib, hash_type)(
                message.encode(encoding=encoding)
            ).hexdigest(), encoding=encoding)


@pytest.mark.parametrize(
    "hash_machine, bytes_message, hash_type, encoding, security", [
        (hash_machines[i], bytes_messages[i], hash_types[i], encodings[i], securities[i]) for i in range(
            len(hash_machines))])
def test_bytes_hash(
        hash_machine,
        bytes_message,
        hash_type,
        encoding,
        security):
    """Tests single bytes hashing for all combinations of hash and encoding types
    """
    if hash_type == 'sha256' and encoding == 'utf_8' and security:  # Genuinely activated security standards
        assert hash_machine.hash(bytes_message) == bytes(
            getattr(hashlib, hash_type)(
                bytes('\x00' + message, encoding=encoding)
            ).hexdigest(), encoding=encoding)
    else:
        assert hash_machine.hash(bytes_message) == bytes(
            getattr(hashlib, hash_type)(
                bytes(message, encoding=encoding)
            ).hexdigest(), encoding=encoding)


@pytest.mark.parametrize(
    "hash_machine, bytearray_message, hash_type, encoding, security",
    [
        (hash_machines[i],
         bytearray_messages[i],
         hash_types[i],
         encodings[i],
         securities[i]) for i in range(
            len(hash_machines))])
def test_bytearray_hash(
        hash_machine,
        bytearray_message,
        hash_type,
        encoding,
        security):
    """Tests single bytearray hashing for all combinations of hash and encoding types
    """
    if hash_type == 'sha256' and encoding == 'utf_8' and security:
        # Genuinely activated security standards
        assert hash_machine.hash(bytearray_message) == bytes(
            getattr(hashlib, hash_type)(
                bytearray('\x00' + message, encoding=encoding)
            ).hexdigest(), encoding=encoding)
    else:
        assert hash_machine.hash(bytearray_message) == bytes(
            getattr(hashlib, hash_type)(
                bytearray(message, encoding=encoding)
            ).hexdigest(), encoding=encoding)


@pytest.mark.parametrize(
    "hash_machine, hash_type, encoding, security", [
        (hash_machines[i], hash_types[i], encodings[i], securities[i]) for i in range(
            len(hash_machines))])
def test_double_bytearray_hash(hash_machine, hash_type, encoding, security):
    """Tests two arguments hashing for all combinations of hash and encoding types
    """
    if hash_type == 'sha256' and encoding == 'utf_8' and security:
        # Genuinely activated security standards
        assert hash_machine.hash(
            bytearray(
                message,
                encoding=encoding),
            bytearray(
                message,
                encoding=encoding)) == bytes(
            getattr(
                hashlib,
                hash_type)(
                bytes(
                    '\x01' +
                    message +
                    '\x01' +
                    message,
                    encoding=encoding)).hexdigest(),
            encoding=encoding)
    else:
        assert hash_machine.hash(
            bytearray(
                message,
                encoding=encoding),
            bytearray(
                message,
                encoding=encoding)) == bytes(
            getattr(
                hashlib,
                hash_type)(
                bytes(
                    message +
                    message,
                    encoding=encoding)).hexdigest(),
            encoding=encoding)


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_empty_multi_hash(hash_machine):
    """Tests multi_hash with edge case argument for all combinations of hash and encoding types
    """
    assert hash_machine.multi_hash([], start='anything') is None


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_with_one_arg(hash_machine):
    """Tests that multi_hash output with one arg coincides with that of hash
    for all possible combinations of hash and encoding types
    """
    assert hash_machine.multi_hash(
        [(+1, hash_machine.hash(message))], start=0) == hash_machine.hash(message)


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_with_two_args(hash_machine):
    """Tests that multi_hash output with two args coincides with that of hash
    for all possible combinations of hash and encoding types
    """
    multi_hash = hash_machine.multi_hash
    hash = hash_machine.hash
    if hash_machine.SECURITY:
        # Genuinely activated security standards
        assert multi_hash(
            [
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=0) == multi_hash(
            [
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=1) == hash(
            bytes(
                message,
                encoding=hash_machine.ENCODING),
            bytes(
                message,
                encoding=hash_machine.ENCODING))  # integrates security prefices
    else:
        assert multi_hash([(+1,
                            bytes(message,
                                  encoding=hash_machine.ENCODING)),
                           (-1,
                            bytes(message,
                                  encoding=hash_machine.ENCODING))],
                          start=0) == multi_hash([(+1,
                                                   bytes(message,
                                                         encoding=hash_machine.ENCODING)),
                                                  (-1,
                                                   bytes(message,
                                                         encoding=hash_machine.ENCODING))],
                                                 start=1) == hash(message + message)


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_with_three_args_first_case(hash_machine):
    """Tests first case output of multi_hash with three args for all possible
    combinations of hash and encoding types
    """
    multi_hash = hash_machine.multi_hash
    hash = hash_machine.hash
    if hash_machine.SECURITY:
        # Genuinely activated security standards
        assert multi_hash(
            signed_hashes=[
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                ('_anything_',
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=0) == multi_hash(
            signed_hashes=[
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                ('_anything_',
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=1) == hash(
            hash(
                bytes(
                    message,
                    encoding=hash_machine.ENCODING),
                bytes(
                    message,
                    encoding=hash_machine.ENCODING)),
            bytes(
                message,
                encoding=hash_machine.ENCODING))  # integrates security prefices
    else:
        assert multi_hash(
            signed_hashes=[
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                ('_anything_',
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=0) == multi_hash(
            signed_hashes=[
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                ('_anything_',
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=1) == hash(
            hash(
                message + message),
            bytes(
                message,
                encoding=hash_machine.ENCODING))


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_with_three_args_second_case(hash_machine):
    """Tests second case output of multi_hash with three args for all possible
    combinations of hash and encoding types
    """
    multi_hash = hash_machine.multi_hash
    hash = hash_machine.hash
    if hash_machine.SECURITY:
        # Genuinely activated security standards
        assert multi_hash(
            signed_hashes=[
                ('_anything_',
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     hash_machine.ENCODING))],
            start=2) == multi_hash(
            signed_hashes=[
                ('_anything_',
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     hash_machine.ENCODING))],
            start=1) == hash(
            bytes(
                message,
                hash_machine.ENCODING),
            hash(
                bytes(
                    message,
                    hash_machine.ENCODING),
                bytes(
                    message,
                    hash_machine.ENCODING)))  # integrates security prefices
    else:
        assert multi_hash(
            signed_hashes=[
                ('_anything_',
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     hash_machine.ENCODING))],
            start=2) == multi_hash(
            signed_hashes=[
                ('_anything_',
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     hash_machine.ENCODING))],
            start=1) == hash(
            bytes(
                message,
                hash_machine.ENCODING),
            hash(
                message + message))


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_four_args_first_edge_case(hash_machine):
    """Tests first edge case of multi_hash with four args
    """
    multi_hash = hash_machine.multi_hash
    hash = hash_machine.hash
    if hash_machine.SECURITY:
        # Genuinely activated security standards
        assert multi_hash(
            signed_hashes=[
                (+1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                ('_anything_',
                 bytes(message, encoding=hash_machine.ENCODING))],
            start=0) == hash(
            hash(
                hash(
                    bytes(message, encoding=hash_machine.ENCODING),
                    bytes(message, encoding=hash_machine.ENCODING)),
                bytes(message, encoding=hash_machine.ENCODING)),
            bytes(message, encoding=hash_machine.ENCODING))  # integrates security prefices
    else:
        assert multi_hash(
            signed_hashes=[
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                ('_anything_',
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=0) == hash(
            hash(
                hash(
                    message + message),
                bytes(
                    message,
                    encoding=hash_machine.ENCODING)),
            bytes(
                message,
                encoding=hash_machine.ENCODING))


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_four_args_second_edge_case(hash_machine):
    """Tests second edge case of multi_hash with four args
    """
    multi_hash = hash_machine.multi_hash
    hash = hash_machine.hash
    if hash_machine.SECURITY:
        # Genuinely activated security standards
        assert multi_hash(
            signed_hashes=[
                ('_anything_',
                 bytes(message, encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(message, encoding=hash_machine.ENCODING))],
            start=3) == hash(
            bytes(message, encoding=hash_machine.ENCODING),
            hash(
                bytes(message, encoding=hash_machine.ENCODING),
                hash(
                    bytes(message, encoding=hash_machine.ENCODING),
                    bytes(message, encoding=hash_machine.ENCODING))))  # integrates security prefices
    else:
        assert multi_hash(
            signed_hashes=[
                ('_anything_',
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=3) == hash(
            bytes(
                message,
                encoding=hash_machine.ENCODING),
            hash(
                bytes(
                    message,
                    encoding=hash_machine.ENCODING),
                hash(
                    message + message)))


@pytest.mark.parametrize('hash_machine', hash_machines)
def test_multi_hash_with_four_args(hash_machine):
    """Tests multi_hash for some non-edge case of four args
    """
    multi_hash = hash_machine.multi_hash
    hash = hash_machine.hash
    if hash_machine.SECURITY:
        # Genuinely activated security standards
        assert multi_hash(
            signed_hashes=[
                (+1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(message, encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(message, encoding=hash_machine.ENCODING))],
            start=1) == hash(
            hash(
                bytes(message, encoding=hash_machine.ENCODING),
                hash(
                    bytes(message, encoding=hash_machine.ENCODING),
                    bytes(message, encoding=hash_machine.ENCODING))),
            bytes(message, encoding=hash_machine.ENCODING))  # integrates security prefices
    else:
        assert multi_hash(
            signed_hashes=[
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (+1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING)),
                (-1,
                 bytes(
                     message,
                     encoding=hash_machine.ENCODING))],
            start=1) == hash(
            hash(
                bytes(
                    message,
                    encoding=hash_machine.ENCODING),
                hash(
                    message + message)),
            bytes(
                message,
                encoding=hash_machine.ENCODING))



# -*- coding: utf-8 -*-
#!/usr/bin/env python
from setuptools import setup, find_packages
import pymerkle

with open("README.md", "r") as f:
    long_description = f.read()

python_requires = '>=3.6'

setup(
    name="pymerkle",
    version=pymerkle.__version__,
    author="Foteinos Mergoupis",
    author_email="foteinosmerg@gmail.com",
    description="A Python library for constructing Merkle Trees and validating Log Proofs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/FoteinosMerg/pymerkle",
    packages=find_packages(),
    install_requires=open("requirements.txt").readlines(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Security :: Cryptography",
        "Topic :: Software Development :: Libraries :: Python Modules"],
    license="License :: OSI Approved :: MIT License",
    keywords="merkle proof audit consistency log security encryption")



"""
Provides the main class for `proof` objects and related functionalities
"""
import uuid
import time
import json
from .serializers import ProofSerializer
from .utils import stringify_path


# -------------------------------- Main class --------------------------------


class Proof(object):
    """Base class for the ``proof`` object

    Basic purpose of this class is to organize the provided proof data in a nicely formatted dictionary,
    so that they can be handled by the ``validations`` module.

    :param generation:  Should be ``'SUCCESS'`` or ``'FAILURE'`` plus explanation message, according to whether
                        or not a proof path could indeed be generated by the provider Merkle-tree
    :type generation:   str
    :param provider:    uuid of the provider Merkle-tree
    :type provider:     str
    :param hash_type:   Hash type of the provider Merkle-tree
    :type hash_type:    str
    :param security:    Security mode of the provider Merkle-tree
    :type security:     bool
    :param encoding:    Encoding type of the provider Merkle-tree
    :type encoding:     str
    :param proof_index: Path index where the validation procedure should start from
    :type proof_index:  int
    :param proof_path:  Path of signed hashes
    :type proof_path:   list of (+1/-1, bytes)

    .. note:: Required Merkle-tree parameters (``hash_type``, ``security`` and ``encoding``) are necessary for
              proof validation (cf. the ``pymerkle.validations`` module)

    Instead of providing the above arguments corresponding to `*args`, a ``proof`` object may also be constructed
    in the following ways by employing `**kwargs` in order to load the JSON string of a given proof ``p``:

    >>> from pymerkle.proof import proof
    >>> q = proof(from_json=p.JSONstring())
    >>> r = proof(from_dict=json.loads(p.JSONstring()))

    .. note:: Constructing proofs in the above ways is a genuine *replication*, since the constructed
              proofs ``q`` and ``r`` have the same *uuid* and *timestamps* as ``p``

    :ivar header:                 (*dict*) Contains the keys *uuid*, *generation*, *timestamp*, *creation_moment*,
                                  *provider*, *hash_type*, *encoding*, *security* and *status* (see below)
    :ivar header.uuid:            (*str*) uuid of the proof (time-based)
    :ivar header.generation:      (*bool*) ``True`` or ``False`` according to whether or not a proof path could
                                  indeed be generated by the provider Merkle-tree
    :ivar header.timestamp:       (*str*) Creation moment (msecs) from the start of time
    :ivar header.creation_moment: (*str*) Creation moment in human readable form
    :ivar header.provider:        (*str*) uuid of the provider Merkle-tree
    :ivar header.hash_type:       (*str*) Hash type of the provider Merkle-tree
    :ivar header.encoding:        (*str*) Encoding type of the provider Merkle-tree
    :ivar header.security:        (*bool*) Security mode of the provider Merkle-tree
    :ivar header.status:          (*bool*) ``True`` resp. ``False`` if the proof was found to be *valid*, resp. *invalid*
                                  after the last validation. If no validation has yet been performed, then it
                                  is ``None``.
    :ivar body:                   (*dict*) Contains the keys *proof_index*, *proof_path* (see below)
    :ivar body.proof_index:       (*int*) See the homonymous argument of the constructor
    :ivar body.proof_path:        (*list of (+1/-1, bytes)*) See the homonymous argument of the constructor
    """

    def __init__(self, *args, **kwargs):
        if args:                                    # Assuming positional arguments by default
            self.header = {
                'uuid': str(uuid.uuid1()),      # Time based proof idW
                'timestamp': int(time.time()),
                'creation_moment': time.ctime(),
                'generation': args[0],
                'provider': args[1],
                'hash_type': args[2],
                'security': args[3],
                'encoding': args[4],
                'status': None                  # Will change to True or False after validation
            }

            self.body = {
                'proof_index': args[5],
                'proof_path': args[6]
            }
        else:
            if kwargs.get('from_dict'):             # Importing proof from dict
                self.header = kwargs.get('from_dict')['header']
                self.body = kwargs.get('from_dict')['body']
            elif kwargs.get('from_json'):           # Importing proof from JSON text
                proof_dict = json.loads(kwargs.get('from_json'))
                self.header = proof_dict['header']
                self.body = proof_dict['body']
            else:                                   # Standard creation of a proof
                self.header = {
                    'uuid': str(uuid.uuid1()),      # Time based proof idW
                    'timestamp': int(time.time()),
                    'creation_moment': time.ctime(),
                    'generation': kwargs.get('generation'),
                    'provider': kwargs.get('provider'),
                    'hash_type': kwargs.get('hash_type'),
                    'security': kwargs.get('security'),
                    'encoding': kwargs.get('encoding'),
                    'status': None                  # Will change to True or False after validation
                }

                self.body = {
                    'proof_index': kwargs.get("proof_index"),
                    'proof_path': kwargs.get("proof_path")
                }

    def __repr__(self):
        """Overrides the default implementation.

        Sole purpose of this function is to easy print info about a proof by just invoking it at console.

        .. warning:: Contrary to convention, the output of this implementation is *not* insertible to the ``eval`` function
        """

        return '\n    ----------------------------------- PROOF ------------------------------------\
                \n\
                \n    uuid        : {uuid}\
                \n\
                \n    generation  : {generation}\
                \n\
                \n    timestamp   : {timestamp} ({creation_moment})\
                \n    provider    : {provider}\
                \n\
                \n    hash-type   : {hash_type}\
                \n    encoding    : {encoding}\
                \n    security    : {security}\
                \n\
                \n    proof-index : {proof_index}\
                \n    proof-path  :\
                \n    {proof_path}\
                \n\
                \n    status      : {status}\
                \n\
                \n    -------------------------------- END OF PROOF --------------------------------\
                \n'.format(
            uuid=self.header['uuid'],
            generation=self.header['generation'],
            timestamp=self.header['timestamp'],
            creation_moment=self.header['creation_moment'],
            provider=self.header['provider'],
            hash_type=self.header['hash_type'].upper().replace('_', '-'),
            encoding=self.header['encoding'].upper().replace('_', '-'),
            security='ACTIVATED' if self.header['security'] else 'DEACTIVATED',
            proof_index=self.body['proof_index'] if self.body['proof_index'] is not None else '',
            proof_path=stringify_path(
                signed_hashes=self.body['proof_path'], encoding=self.header['encoding']),
            status='UNVALIDATED' if self.header['status'] is None
            else 'VALID' if self.header['status'] is True
            else 'NON VALID')

# ------------------------------- Serialization --------------------------

    def serialize(self):
        """ Returns a JSON entity with the proof's current characteristics as key-value pairs

        :rtype: dict
        """
        serializer = ProofSerializer()
        return serializer.default(self)

    def JSONstring(self):
        """Returns a nicely stringified version of the proof's JSON serialized form

        .. note:: The output of this function is to be passed into the ``print`` function

        :rtype: str
        """
        return json.dumps(
            self,
            cls=ProofSerializer,
            sort_keys=True,
            indent=4)
